from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, Generic, Self, Unpack

from django.db.models import Model
from graphql import DirectiveLocation

from undine.exceptions import (
    GraphQLUnionResolveTypeInvalidValueError,
    GraphQLUnionResolveTypeModelNotFoundError,
    MissingUnionQueryTypeGenericError,
)
from undine.parsers import parse_class_attribute_docstrings
from undine.settings import undine_settings
from undine.typing import TQueryTypes
from undine.utils.graphql.type_registry import get_or_create_graphql_union
from undine.utils.graphql.utils import check_directives
from undine.utils.text import get_docstring

if TYPE_CHECKING:
    from collections.abc import Iterable

    from graphql import GraphQLAbstractType, GraphQLUnionType

    from undine import FilterSet, GQLInfo, OrderSet, QueryType
    from undine.directives import Directive
    from undine.typing import DjangoRequestProtocol, UnionTypeParams

__all__ = [
    "UnionType",
]


class UnionTypeMeta(type):
    """A metaclass that modifies how a `UnionType` is created."""

    # Set in '__new__'
    __query_types_by_model__: dict[type[Model], type[QueryType]]
    __schema_name__: str
    __filterset__: type[FilterSet] | None
    __orderset__: type[OrderSet] | None
    __directives__: list[Directive]
    __extensions__: dict[str, Any]
    __attribute_docstrings__: dict[str, str]

    # Internal use only
    __query_types__: Iterable[type[QueryType]]

    def __new__(
        cls,
        _name: str,
        _bases: tuple[type, ...],
        _attrs: dict[str, Any],
        **kwargs: Unpack[UnionTypeParams],
    ) -> UnionTypeMeta:
        if _name == "UnionType":  # Early return for the `UnionType` class itself.
            return super().__new__(cls, _name, _bases, _attrs)

        try:
            query_types = UnionTypeMeta.__query_types__
            del UnionTypeMeta.__query_types__
        except AttributeError as error:
            raise MissingUnionQueryTypeGenericError(name=_name, cls="UnionType") from error

        union_type = super().__new__(cls, _name, _bases, _attrs)

        # Members should use `__dunder__` names to avoid name collisions with possible `UnionType` names.
        union_type.__query_types_by_model__ = {query_type.__model__: query_type for query_type in query_types}

        union_type.__filterset__ = kwargs.get("filterset")
        if union_type.__filterset__ is not None:
            union_type.__filterset__.__add_to_union_type__(union_type)  # type: ignore[arg-type]

        union_type.__orderset__ = kwargs.get("orderset")
        if union_type.__orderset__ is not None:
            union_type.__orderset__.__add_to_union_type__(union_type)  # type: ignore[arg-type]

        union_type.__schema_name__ = kwargs.get("schema_name", _name)
        union_type.__directives__ = kwargs.get("directives", [])
        union_type.__extensions__ = kwargs.get("extensions", {})
        union_type.__attribute_docstrings__ = parse_class_attribute_docstrings(union_type)

        check_directives(union_type.__directives__, location=DirectiveLocation.UNION)
        union_type.__extensions__[undine_settings.UNION_TYPE_EXTENSIONS_KEY] = union_type

        return union_type

    def __str__(cls) -> str:
        return undine_settings.SDL_PRINTER.print_union_type(cls.__union_type__())

    def __getitem__(cls, query_types: tuple[type[QueryType], ...]) -> type[UnionType[*TQueryTypes]]:
        # Note that this should be cleaned up in '__new__',
        # but is not if an error occurs in the class body of the defined 'UnionType'!
        UnionTypeMeta.__query_types__ = query_types
        return cls  # type: ignore[return-value]

    def __resolve_type__(cls, value: Any, info: GQLInfo, abstract_type: GraphQLAbstractType) -> str:
        if not isinstance(value, Model):
            raise GraphQLUnionResolveTypeInvalidValueError(name=cls.__schema_name__, value=value)

        model = value.__class__
        query_type = cls.__query_types_by_model__.get(model)
        if query_type is None:
            raise GraphQLUnionResolveTypeModelNotFoundError(name=cls.__schema_name__, model=model)

        return query_type.__schema_name__

    def __union_type__(cls) -> GraphQLUnionType:
        return get_or_create_graphql_union(
            name=cls.__schema_name__,
            types=[query_type.__output_type__() for query_type in cls.__query_types_by_model__.values()],
            resolve_type=cls.__resolve_type__,
            description=get_docstring(cls),
            extensions=cls.__extensions__,
        )

    def __add_directive__(cls, directive: Directive, /) -> Self:
        """Add a directive to this union."""
        check_directives([directive], location=DirectiveLocation.UNION)
        cls.__directives__.append(directive)
        return cls


class UnionType(Generic[*TQueryTypes], metaclass=UnionTypeMeta):
    """
    A class for creating a GraphQL Union based on two or more `QueryTypes`.

    Must set the `QueryTypes` this `UnionType` contains using the generic type argument.

    The following parameters can be passed in the class definition:

    `schema_name: str = <class name>`
        Override name for `UnionType` in the GraphQL schema.

    `directives: list[Directive] = []`
        `Directives` to add to the created `UnionType`.

    `extensions: dict[str, Any] = {}`
        GraphQL extensions for the created `UnionType`.

    >>> class TaskType(QueryType[Task]): ...
    >>> class ProjectType(QueryType[Project]): ...
    >>>
    >>> class Commentable(UnionType[TaskType, ProjectType]): ...
    """

    # Set in metaclass
    __query_types_by_model__: ClassVar[dict[type[Model], type[QueryType]]]
    __schema_name__: ClassVar[str]
    __filterset__: ClassVar[type[FilterSet] | None]
    __orderset__: ClassVar[type[OrderSet] | None]
    __directives__: ClassVar[list[Directive]]
    __extensions__: ClassVar[dict[str, Any]]
    __attribute_docstrings__: ClassVar[dict[str, str]]

    @classmethod
    def __is_visible__(cls, request: DjangoRequestProtocol) -> bool:
        """
        Determine if the given `UnionType` is visible in the schema.
        Experimental, requires `EXPERIMENTAL_VISIBILITY_CHECKS` to be enabled.
        """
        return True



from __future__ import annotations

from django.core.files import File
from django.core.files.uploadedfile import UploadedFile
from django.db.models.fields.files import FieldFile

from undine.utils.validators import validate_file_upload, validate_file_url

from ._definition import ScalarType

__all__ = [
    "GraphQLFile",
    "file_scalar",
]


file_scalar: ScalarType[UploadedFile, str] = ScalarType(
    name="File",
    description="Represents any kind of file.",
)

GraphQLFile = file_scalar.as_graphql_scalar()


@file_scalar.parse.register
def _(value: UploadedFile) -> UploadedFile:
    validate_file_upload(value)
    return value


@file_scalar.serialize.register
def _(value: FieldFile) -> str:
    return value.url


@file_scalar.serialize.register
def _(value: File) -> str:
    return value.name or ""


@file_scalar.serialize.register
def _(value: str) -> str:
    validate_file_url(value)
    return value



from __future__ import annotations

from contextlib import suppress
from typing import Any

from django.contrib.contenttypes.fields import GenericRelation
from django.db.models import (
    BinaryField,
    BooleanField,
    CharField,
    DateField,
    DateTimeField,
    DecimalField,
    DurationField,
    EmailField,
    FileField,
    FilePathField,
    FloatField,
    ForeignKey,
    GenericIPAddressField,
    ImageField,
    IntegerField,
    IPAddressField,
    JSONField,
    ManyToManyField,
    ManyToManyRel,
    ManyToOneRel,
    NullBooleanField,
    OneToOneField,
    OneToOneRel,
    SlugField,
    TextField,
    TimeField,
    URLField,
    UUIDField,
)

from undine.converters.definitions import convert_to_bad_lookups
from undine.typing import SupportsLookup
from undine.utils.model_utils import get_many_to_many_through_field


@convert_to_bad_lookups.register
def _(_: SupportsLookup, **kwargs: Any) -> set[str]:
    return set()  # By default, don't exclude any lookups


@convert_to_bad_lookups.register
def _(_: BooleanField, **kwargs: Any) -> set[str]:
    return {
        "contains",
        "endswith",
        "gt",
        "gte",
        "icontains",
        "iendswith",
        "iexact",
        "in",
        "isnull",
        "istartswith",
        "lt",
        "lte",
        "range",
        "startswith",
    }


@convert_to_bad_lookups.register
def _(
    field: (
        CharField
        | EmailField
        | FilePathField
        | IPAddressField
        | GenericIPAddressField
        | SlugField
        | TextField
        | URLField
    ),
    **kwargs: Any,
) -> set[str]:
    lookups = {
        "gt",
        "gte",
        "lt",
        "lte",
        "range",
    }
    if not field.null:
        lookups.add("isnull")
    return lookups


@convert_to_bad_lookups.register
def _(field: DateField, **kwargs: Any) -> set[str]:
    lookups = {
        "contained_by",
        "contains",
        "endswith",
        "icontains",
        "iendswith",
        "iexact",
        "istartswith",
        "startswith",
    }
    if not field.null:
        lookups.add("isnull")
    return lookups


@convert_to_bad_lookups.register
def _(field: DateTimeField, **kwargs: Any) -> set[str]:
    lookups = {
        "contained_by",
        "contains",
        "endswith",
        "icontains",
        "iendswith",
        "iexact",
        "istartswith",
        "startswith",
    }
    if not field.null:
        lookups.add("isnull")
    return lookups


@convert_to_bad_lookups.register
def _(field: DurationField, **kwargs: Any) -> set[str]:
    lookups = {
        "contains",
        "endswith",
        "icontains",
        "iendswith",
        "iexact",
        "istartswith",
        "startswith",
    }
    if not field.null:
        lookups.add("isnull")
    return lookups


@convert_to_bad_lookups.register
def _(field: IntegerField | FloatField | DecimalField, **kwargs: Any) -> set[str]:
    lookups = {
        "contained_by",
        "icontains",
        "iendswith",
        "iexact",
        "istartswith",
    }
    if not field.null:
        lookups.add("isnull")
    return lookups


@convert_to_bad_lookups.register
def _(_: NullBooleanField, **kwargs: Any) -> set[str]:
    return {
        "contains",
        "endswith",
        "gt",
        "gte",
        "icontains",
        "iendswith",
        "iexact",
        "istartswith",
        "lt",
        "lte",
        "range",
        "startswith",
    }


@convert_to_bad_lookups.register
def _(field: TimeField, **kwargs: Any) -> set[str]:
    lookups = {
        "icontains",
        "iendswith",
        "iexact",
        "istartswith",
    }
    if not field.null:
        lookups.add("isnull")
    return lookups


@convert_to_bad_lookups.register
def _(field: BinaryField, **kwargs: Any) -> set[str]:
    lookups = {
        "gt",
        "gte",
        "icontains",
        "iendswith",
        "iexact",
        "istartswith",
        "lt",
        "lte",
    }
    if not field.null:
        lookups.add("isnull")
    return lookups


@convert_to_bad_lookups.register
def _(field: UUIDField, **kwargs: Any) -> set[str]:
    lookups = {
        "gt",
        "gte",
        "icontains",
        "iendswith",
        "iexact",
        "istartswith",
        "lt",
        "lte",
        "range",
    }
    if not field.null:
        lookups.add("isnull")
    return lookups


@convert_to_bad_lookups.register
def _(field: JSONField, **kwargs: Any) -> set[str]:
    lookups = {
        "endswith",
        "gt",
        "gte",
        "icontains",
        "iendswith",
        "iexact",
        "istartswith",
        "lt",
        "lte",
        "range",
        "startswith",
    }
    if not field.null:
        lookups.add("isnull")
    return lookups


@convert_to_bad_lookups.register
def _(field: FileField | ImageField, **kwargs: Any) -> set[str]:
    lookups: set[str] = {
        "gt",
        "gte",
        "icontains",
        "iendswith",
        "iexact",
        "istartswith",
        "lt",
        "lte",
        "range",
    }
    if not field.null:
        lookups.add("isnull")
    return lookups


@convert_to_bad_lookups.register
def _(field: OneToOneField | OneToOneRel | ForeignKey | ManyToOneRel, **kwargs: Any) -> set[str]:
    lookups: set[str] = set()
    if not field.null:
        lookups.add("isnull")
    return lookups


@convert_to_bad_lookups.register
def _(field: ManyToManyField | ManyToManyRel, **kwargs: Any) -> set[str]:
    through_field = get_many_to_many_through_field(field)

    # For many-to-many fields, any lookups are done to the corresponding field on the through-model.
    # Remove any lookups from the many-to-many fields' lookups that are not on the through-model field,
    # as well as any bad lookups registered for the the through-model field's type.
    bad_lookups = set(field.get_lookups()) - set(through_field.get_lookups())
    bad_lookups |= convert_to_bad_lookups(through_field)

    return bad_lookups


with suppress(ImportError):
    from django.contrib.postgres.fields import ArrayField, HStoreField, RangeField

    @convert_to_bad_lookups.register
    def _(field: ArrayField, **kwargs: Any) -> set[str]:
        lookups = {
            "endswith",
            "gt",
            "gte",
            "icontains",
            "iendswith",
            "iexact",
            "in",
            "istartswith",
            "lt",
            "lte",
            "range",
            "startswith",
        }
        if not field.null:
            lookups.add("isnull")
        return lookups

    @convert_to_bad_lookups.register
    def _(field: HStoreField, **kwargs: Any) -> set[str]:
        lookups = {
            "endswith",
            "gt",
            "gte",
            "icontains",
            "iendswith",
            "iexact",
            "in",
            "istartswith",
            "lt",
            "lte",
            "range",
            "startswith",
        }
        if not field.null:
            lookups.add("isnull")
        return lookups

    @convert_to_bad_lookups.register
    def _(field: RangeField, **kwargs: Any) -> set[str]:
        lookups = {
            "adjacent_to",
            "contained_by",
            "contains",
            "fully_gt",
            "fully_lt",
            "gt",
            "gte",
            "icontains",
            "iendswith",
            "iexact",
            "in",
            "istartswith",
            "lt",
            "lte",
            "not_gt",
            "not_lt",
            "overlap",
            "range",
        }
        if not field.null:
            lookups.add("isnull")
        return lookups


@convert_to_bad_lookups.register
def _(_: GenericRelation, **kwargs: Any) -> set[str]:
    return set()



from __future__ import annotations

from importlib import import_module
from pathlib import Path

from django.apps import AppConfig
from graphql import GraphQLWrappingType


class UndineConfig(AppConfig):
    name = "undine"
    label = "undine"
    verbose_name = "undine"

    def ready(self) -> None:
        self.patch_graphql_wrapping_object()
        self.register_additional_types()
        self.patch_introspection_types()
        self.register_converters()
        self.maybe_disable_did_you_mean()
        self.patch_debug_toolbar_if_installed()

    def patch_graphql_wrapping_object(self) -> None:
        """Set wrapping types to compare their wrapped types."""

        def wrapping_eq(self_: GraphQLWrappingType, other: object) -> bool:
            if not isinstance(other, type(self_)):
                return NotImplemented
            return self_.of_type == other.of_type

        GraphQLWrappingType.__eq__ = wrapping_eq  # type: ignore[method-assign,assignment]

    def register_additional_types(self) -> None:
        from undine.utils.graphql.type_registry import register_builtins  # noqa: PLC0415

        register_builtins()

    def patch_introspection_types(self) -> None:
        from undine.settings import undine_settings  # noqa: PLC0415

        if not undine_settings.EXPERIMENTAL_VISIBILITY_CHECKS:
            return

        from undine.utils.graphql.introspection import patch_introspection_schema  # noqa: PLC0415

        patch_introspection_schema()

    def register_converters(self) -> None:
        """Import all converter implementation modules to register the implementations to the converters."""
        import undine.converters  # noqa: PLC0415

        converter_dir = Path(undine.converters.__file__).resolve().parent
        lib_root = converter_dir.parent.parent

        for file in converter_dir.glob("impl/*.py"):
            import_path = file.relative_to(lib_root).as_posix().replace("/", ".").removesuffix(".py")
            import_module(import_path)

    def maybe_disable_did_you_mean(self) -> None:
        """Disable the 'did you mean' suggestions on error messages if `DISABLE_DID_YOU_MEAN` is True."""
        from graphql.pyutils import did_you_mean  # noqa: PLC0415

        from undine.settings import undine_settings  # noqa: PLC0415

        # See: https://github.com/graphql-python/graphql-core/issues/97#issuecomment-642967670
        if not undine_settings.ALLOW_DID_YOU_MEAN_SUGGESTIONS:
            did_you_mean.__globals__["MAX_LENGTH"] = 0

    def patch_debug_toolbar_if_installed(self) -> None:
        """Patch `django-debug-toolbar` to work with Undine if it's installed."""
        try:
            from undine.integrations.debug_toolbar import monkeypatch_middleware  # noqa: PLC0415

        # `ImportError` raised if debug-toolbar is not installed in Python environment
        # `RuntimeError` raised if debug-toolbar is not installed in Django project
        except (ImportError, RuntimeError):
            return

        monkeypatch_middleware()



from __future__ import annotations

from types import FunctionType
from typing import Any

from django.contrib.contenttypes.fields import GenericForeignKey
from django.db.models import Model

from undine import Input, MutationType
from undine.converters import is_input_only
from undine.dataclasses import LazyLambda, TypeRef
from undine.exceptions import ModelFieldError
from undine.typing import ModelField
from undine.utils.model_utils import get_model_field


@is_input_only.register
def _(_: ModelField, **kwargs: Any) -> bool:
    return False


@is_input_only.register
def _(_: TypeRef | FunctionType, **kwargs: Any) -> bool:
    caller: Input = kwargs["caller"]

    try:
        field = get_model_field(model=caller.mutation_type.__model__, lookup=caller.field_name)
    except ModelFieldError:
        return True
    return is_input_only(field)


@is_input_only.register
def _(_: LazyLambda, **kwargs: Any) -> bool:
    return False


@is_input_only.register
def _(_: type[MutationType], **kwargs: Any) -> bool:
    return False


@is_input_only.register  # Required for Django<5.1
def _(_: GenericForeignKey, **kwargs: Any) -> bool:
    return False


@is_input_only.register
def _(_: type[Model], **kwargs: Any) -> bool:
    caller: Input = kwargs["caller"]

    try:
        field = get_model_field(model=caller.mutation_type.__model__, lookup=caller.field_name)
    except ModelFieldError:
        return True

    return is_input_only(field, **kwargs)



from __future__ import annotations

import string
from typing import Any

from django.core.exceptions import ValidationError
from graphql import GraphQLError, parse

from undine.execution import _validate
from undine.settings import undine_settings

__all__ = [
    "validate_document",
    "validate_document_id",
]


VALID_CHARS = frozenset(string.ascii_letters + string.digits + ":-._~")


def validate_document_id(value: Any) -> None:
    """Validate the document id of a persisted document."""
    if not isinstance(value, str):
        msg = "Document ID must be a string"
        raise ValidationError(msg)

    invalid_chars: set[str] = {c for c in value if c not in VALID_CHARS}
    if invalid_chars:
        msg = f"Document ID contains invalid characters: {' '.join(sorted(invalid_chars))}"
        raise ValidationError(msg)


def validate_document(value: Any) -> None:
    """Validate the document of a persisted document."""
    if not isinstance(value, str):
        msg = "Document must be a string"
        raise ValidationError(msg)

    try:
        document = parse(
            source=value,
            no_location=undine_settings.NO_ERROR_LOCATION,
            max_tokens=undine_settings.MAX_TOKENS,
        )
    except GraphQLError as parse_error:
        raise ValidationError(parse_error.message) from parse_error

    validation_errors = _validate(document=document)
    if validation_errors:
        raise ValidationError([error.message for error in validation_errors])



from __future__ import annotations

from typing import Any

from django.core.management import BaseCommand

from undine.settings import undine_settings


class Command(BaseCommand):
    help = "Print the GraphQL schema to stdout."

    def handle(self, *args: Any, **options: Any) -> None:
        self.stdout.write(undine_settings.SDL_PRINTER.print_schema(undine_settings.SCHEMA))



from __future__ import annotations

import functools
import itertools
import operator
import operator as op
from collections import defaultdict
from collections.abc import Iterable
from functools import reduce
from typing import TYPE_CHECKING, Any, ClassVar, Generic, Self, Unpack

from django.db.models import Model, Q
from django.db.models.constants import LOOKUP_SEP
from graphql import DirectiveLocation, GraphQLInputField, GraphQLInputType, Undefined

from undine.converters import (
    convert_to_description,
    convert_to_filter_lookups,
    convert_to_filter_ref,
    convert_to_filter_resolver,
    convert_to_graphql_type,
)
from undine.dataclasses import FilterResults, LookupRef, MaybeManyOrNonNull
from undine.exceptions import (
    EmptyFilterResult,
    MismatchingModelError,
    MissingModelGenericError,
    NotCompatibleWithError,
    QueryTypeRequiresSingleModelError,
    UnionTypeModelsDifferentError,
    UnionTypeRequiresMultipleModelsError,
)
from undine.parsers import parse_class_attribute_docstrings
from undine.settings import undine_settings
from undine.typing import ManyMatch, TModels
from undine.utils.graphql.type_registry import get_or_create_graphql_input_object_type
from undine.utils.graphql.utils import check_directives
from undine.utils.model_utils import get_model_field, get_model_fields_for_graphql, is_to_many, lookup_to_display_name
from undine.utils.reflection import (
    FunctionEqualityWrapper,
    cache_signature_if_function,
    get_members,
    get_wrapped_func,
    is_subclass,
)
from undine.utils.text import dotpath, get_docstring, to_schema_name

if TYPE_CHECKING:
    from collections.abc import Container, Iterable

    from django.db.models import Model, QuerySet
    from graphql import GraphQLFieldResolver, GraphQLInputObjectType, GraphQLInputType

    from undine import QueryType, UnionType
    from undine.directives import Directive
    from undine.typing import (
        DjangoExpression,
        DjangoRequestProtocol,
        FilterAliasesFunc,
        FilterParams,
        FilterSetParams,
        GQLInfo,
        T,
        TModel,
        VisibilityFunc,
    )

__all__ = [
    "Filter",
    "FilterSet",
]


class FilterSetMeta(type):
    """A metaclass that modifies how a `FilterSet` is created."""

    # Set in '__new__'
    __models__: tuple[type[Model], ...]
    __filter_map__: dict[str, Filter]
    __schema_name__: str
    __directives__: list[Directive]
    __extensions__: dict[str, Any]
    __attribute_docstrings__: dict[str, str]

    def __new__(
        cls,
        _name: str,
        _bases: tuple[type, ...],
        _attrs: dict[str, Any],
        **kwargs: Unpack[FilterSetParams],
    ) -> FilterSetMeta:
        if _name == "FilterSet":  # Early return for the `FilterSet` class itself.
            return super().__new__(cls, _name, _bases, _attrs)

        try:
            models = FilterSetMeta.__models__
            del FilterSetMeta.__models__
        except AttributeError as error:
            raise MissingModelGenericError(name=_name, cls="FilterSet") from error

        auto = kwargs.get("auto", undine_settings.AUTOGENERATION)
        exclude = set(kwargs.get("exclude", []))
        if auto:
            exclude |= set(_attrs)

            if len(models) == 1:
                _attrs |= get_filters_for_model(models[0], exclude=exclude)
            else:
                _attrs |= get_filters_for_models(models, exclude=exclude)

        filterset = super().__new__(cls, _name, _bases, _attrs)

        # Members should use `__dunder__` names to avoid name collisions with possible `Filter` names.
        filterset.__models__ = models
        filterset.__filter_map__ = get_members(filterset, Filter)
        filterset.__schema_name__ = kwargs.get("schema_name", _name)
        filterset.__directives__ = kwargs.get("directives", [])
        filterset.__extensions__ = kwargs.get("extensions", {})
        filterset.__extensions__[undine_settings.FILTERSET_EXTENSIONS_KEY] = filterset
        filterset.__attribute_docstrings__ = parse_class_attribute_docstrings(filterset)

        check_directives(filterset.__directives__, location=DirectiveLocation.INPUT_OBJECT)

        for name, filter_ in filterset.__filter_map__.items():
            filter_.__connect__(filterset, name)  # type: ignore[arg-type]

        return filterset

    def __str__(cls) -> str:
        return undine_settings.SDL_PRINTER.print_input_object_type(cls.__input_type__())

    def __getitem__(cls, models: type[type[TModel]] | tuple[type[TModel], ...]) -> type[FilterSet[*TModels]]:
        # Note that this should be cleaned up in '__new__',
        # but is not if an error occurs in the class body of the defined 'FilterSet'!
        FilterSetMeta.__models__ = models if isinstance(models, tuple) else (models,)
        return cls  # type: ignore[return-value]

    def __call__(cls, ref: T) -> T:
        """
        Allow adding this FilterSet to a QueryType using a decorator syntax

        >>> class TaskFilterSet(FilterSet[Task]): ...
        >>>
        >>> @TaskFilterSet
        >>> class TaskType(QueryType[Task]): ...
        """
        from undine import QueryType, UnionType  # noqa: PLC0415

        if is_subclass(ref, QueryType):
            cls.__add_to_query_type__(ref)
        elif is_subclass(ref, UnionType):
            cls.__add_to_union_type__(ref)
        else:
            raise NotCompatibleWithError(obj=cls, other=ref)

        return ref

    def __build__(cls, filter_data: dict[str, Any], info: GQLInfo) -> FilterResults:
        """
        Build a list of 'Q' expression from the given filter data to apply to the queryset.
        Also indicate if 'queryset.distinct()' is needed, what aliases are required,
        or if the filtering should result in an empty queryset.

        :param filter_data: The input filter data.
        :param info: The GraphQL resolve info for the request.
        """
        filters: list[Q] = []
        distinct: bool = False
        aliases: dict[str, DjangoExpression] = {}
        none: bool = False
        filter_count: int = 0

        try:
            for filter_name, filter_value in filter_data.items():
                if filter_name == "NOT":
                    if not filter_value:
                        continue

                    results = cls.__build__(filter_value, info)
                    distinct |= results.distinct
                    aliases |= results.aliases
                    filter_count += results.filter_count
                    filters.extend(~frt for frt in results.filters)

                elif filter_name in {"AND", "OR", "XOR"}:
                    if not filter_value:
                        continue

                    results = cls.__build__(filter_value, info)
                    distinct |= results.distinct
                    aliases |= results.aliases
                    filter_count += results.filter_count
                    func = op.and_ if filter_name == "AND" else op.or_ if filter_name == "OR" else op.xor
                    filters.append(reduce(func, results.filters, Q()))

                else:
                    ftr = cls.__filter_map__[filter_name]
                    if filter_value in ftr.empty_values:
                        continue

                    distinct |= ftr.distinct
                    if ftr.aliases_func is not None:
                        aliases |= ftr.aliases_func(ftr, info, value=filter_value)

                    if ftr.many:
                        conditions = (ftr.get_expression(value, info) for value in filter_value)
                        filter_expression = reduce(ftr.match.operator, conditions, Q())
                    else:
                        filter_expression = ftr.get_expression(filter_value, info)

                    filters.append(filter_expression)
                    filter_count += 1

        except EmptyFilterResult:
            none = True

        return FilterResults(filters=filters, aliases=aliases, distinct=distinct, none=none, filter_count=filter_count)

    def __input_type__(cls) -> GraphQLInputObjectType:
        """
        Create the input type to use for the `QueryType` this `FilterSet` is for.

        The fields of the input object type are all the `Filter` instances defined in this `FilterSet`,
        as well as a few special fields (NOT, AND, OR, XOR) for logical operations.
        """

        # Defer creating fields so that logical filters can be added.
        def fields() -> dict[str, GraphQLInputField]:
            inputs = cls.__input_fields__()
            input_field = GraphQLInputField(type_=input_object_type)
            inputs["NOT"] = input_field
            inputs["AND"] = input_field
            inputs["OR"] = input_field
            inputs["XOR"] = input_field
            return inputs

        # Assign to a variable so that `fields()` above can access it.
        input_object_type = get_or_create_graphql_input_object_type(
            name=cls.__schema_name__,
            description=get_docstring(cls),
            fields=FunctionEqualityWrapper(fields, context=cls),
            extensions=cls.__extensions__,
        )
        return input_object_type  # noqa: RET504, RUF100

    def __input_fields__(cls) -> dict[str, GraphQLInputField]:
        """Defer creating fields until all QueryTypes have been registered."""
        return {frt.schema_name: frt.as_graphql_input_field() for frt in cls.__filter_map__.values()}

    def __add_directive__(cls, directive: Directive, /) -> Self:
        """Add a directive to this `FilterSet`."""
        check_directives([directive], location=DirectiveLocation.INPUT_OBJECT)
        cls.__directives__.append(directive)
        return cls

    def __add_to_query_type__(cls, query_type: type[QueryType]) -> None:
        models = cls.__models__
        if len(models) != 1:
            raise QueryTypeRequiresSingleModelError(kind="FilterSet")

        if models[0] is not query_type.__model__:
            raise MismatchingModelError(
                name=cls.__name__,
                given_model=models[0],
                target=query_type.__name__,
                expected_model=query_type.__model__,
            )

        query_type.__filterset__ = cls  # type: ignore[assignment]

    def __add_to_union_type__(cls, union_type: type[UnionType]) -> None:
        models = cls.__models__
        if len(models) == 1:
            raise UnionTypeRequiresMultipleModelsError(kind="FilterSet")

        on_union_type = set(union_type.__query_types_by_model__)
        on_filterset = set(models)

        if on_union_type != on_filterset:
            raise UnionTypeModelsDifferentError(kind="FilterSet")

        union_type.__filterset__ = cls  # type: ignore[assignment]


class FilterSet(Generic[*TModels], metaclass=FilterSetMeta):
    """
    A class for adding filtering for a `QueryType`.

    Must set the Django Model this `FilterSet` is for using the generic type argument.
    Model must match the Model of the `QueryType` this `FilterSet` will be added to.

    The following parameters can be passed in the class definition:

    `auto: bool = <AUTOGENERATION setting>`
        Whether to add `Filter` attributes for all Model fields and their lookups automatically.

    `exclude: list[str] = []`
        Model fields to exclude from the automatically added `Filter` attributes.

    `schema_name: str = <class name>`
        Override the name for the `InputObjectType` for this `FilterSet` in the GraphQL schema.

    `directives: list[Directive] = []`
        `Directives` to add to the created `InputObjectType`.

    `extensions: dict[str, Any] = {}`
        GraphQL extensions for the created `InputObjectType`.

    >>> class TaskFilterSet(FilterSet[Task]): ...
    >>> class TaskQueryType(QueryType[Task], filterset=TaskFilterSet): ...
    """

    # Members should use `__dunder__` names to avoid name collisions with possible `Filter` names.

    # Set in metaclass
    __models__: ClassVar[tuple[type[Model], ...]]
    __filter_map__: ClassVar[dict[str, Filter]]
    __schema_name__: ClassVar[str]
    __directives__: ClassVar[list[Directive]]
    __extensions__: ClassVar[dict[str, Any]]
    __attribute_docstrings__: ClassVar[dict[str, str]]

    @classmethod
    def __filter_queryset__(cls, queryset: QuerySet[TModel], info: GQLInfo) -> QuerySet[TModel]:
        """Filtering that should be done to the queryset after all other filters have been applied."""
        return queryset  # pragma: no cover

    @classmethod
    def __is_visible__(cls, request: DjangoRequestProtocol) -> bool:
        """
        Determine if the given `FilterSet` is visible in the schema.
        Experimental, requires `EXPERIMENTAL_VISIBILITY_CHECKS` to be enabled.
        """
        return True


class Filter:
    """
    A class for defining a possible `FilterSet` input.
    Represents an input field in the GraphQL `InputObjectType` for the `FilterSet` this is added to.

    >>> class TaskFilterSet(FilterSet[Task]):
    ...     name = Filter()
    """

    def __init__(self, ref: Any = None, **kwargs: Unpack[FilterParams]) -> None:
        """
        Create a new `Filter`.

        :param ref: The expression to filter by. Must be convertable by the `convert_to_filter_ref` function.
                    If not provided, use the name of the attribute this is assigned to in the `FilterSet` class.
        :param lookup: The lookup expression to use for the `Filter`.
        :param many: If `True`, the `Filter` will accept a list of values, and filtering will be done by matching
                     all the provided values against the filter condition.
        :param match: Sets the behavior of `many` so that the filter condition will include an item if it
                      matches either "any", "all", or "one_of" of the provided values.
        :param distinct: Does the `Filter` require `queryset.distinct()` to be used?
        :param required: Is the `Filter` is a required input?
        :param empty_values: Values that will be ignored if they are provided as filter values.
        :param description: Description of the `Filter`.
        :param deprecation_reason: If the `Filter` is deprecated, describes the reason for deprecation.
        :param field_name: Name of the field in the Django model. If not provided, use the name of the attribute.
        :param schema_name: Actual name of the `Filter` in the GraphQL schema. Can be used to alias the `Filter`
                            for the schema, or when the desired name is a Python keyword (e.g. `if` or `from`).
        :param directives: GraphQL directives for the `Filter`.
        :param extensions: GraphQL extensions for the `Filter`.
        """
        self.ref: Any = cache_signature_if_function(ref, depth=1)

        self.lookup: str = kwargs.get("lookup", "exact")
        self.many: bool = kwargs.get("many", False)
        self.match: ManyMatch = ManyMatch(kwargs.get("match", ManyMatch.any))
        self.distinct: bool = kwargs.get("distinct", False)
        self.required: bool = kwargs.get("required", False)
        self.empty_values: Container = kwargs.get("empty_values", undine_settings.EMPTY_VALUES)
        self.description: str | None = kwargs.get("description", Undefined)  # type: ignore[assignment]
        self.deprecation_reason: str | None = kwargs.get("deprecation_reason")
        self.field_name: str = kwargs.get("field_name", Undefined)  # type: ignore[assignment]
        self.schema_name: str = kwargs.get("schema_name", Undefined)  # type: ignore[assignment]
        self.directives: list[Directive] = kwargs.get("directives", [])
        self.extensions: dict[str, Any] = kwargs.get("extensions", {})

        check_directives(self.directives, location=DirectiveLocation.INPUT_FIELD_DEFINITION)
        self.extensions[undine_settings.FILTER_EXTENSIONS_KEY] = self

        self.aliases_func: FilterAliasesFunc | None = None
        self.visible_func: VisibilityFunc | None = None

    def __connect__(self, filterset: type[FilterSet], name: str) -> None:
        """Connect this `Filter` to the given `FilterSet` using the given name."""
        self.filterset = filterset
        self.name = name
        self.field_name = self.field_name or name
        self.schema_name = self.schema_name or to_schema_name(name)

        if isinstance(self.ref, str):
            self.field_name = self.ref

        self.ref = convert_to_filter_ref(self.ref, caller=self)

        if self.description is Undefined:
            self.description = self.filterset.__attribute_docstrings__.get(name)
            if self.description is None:
                self.description = convert_to_description(self.ref)

        self.resolver = convert_to_filter_resolver(self.ref, caller=self)

    def __call__(self, ref: GraphQLFieldResolver, /) -> Filter:
        """Called when using as decorator with parenthesis: @Filter()"""
        self.ref = cache_signature_if_function(ref, depth=1)
        return self

    def __repr__(self) -> str:
        return f"<{dotpath(self.__class__)}(ref={self.ref!r}, lookup={self.lookup!r})>"

    def __str__(self) -> str:
        inpt = self.as_graphql_input_field()
        return undine_settings.SDL_PRINTER.print_input_field(self.schema_name, inpt, indent=False)

    def get_expression(self, value: Any, info: GQLInfo) -> Q:
        return self.resolver(self, info, value=value)

    def as_graphql_input_field(self) -> GraphQLInputField:
        return GraphQLInputField(
            type_=self.get_field_type(),
            description=self.description,
            deprecation_reason=self.deprecation_reason,
            out_name=self.name,
            extensions=self.extensions,
        )

    def get_field_type(self) -> GraphQLInputType:
        lookup = LookupRef(ref=self.ref, lookup=self.lookup)
        value = MaybeManyOrNonNull(lookup, many=self.many, nullable=not self.required)
        return convert_to_graphql_type(value, model=self.filterset.__models__[0], is_input=True)  # type: ignore[return-value]

    def aliases(self, func: FilterAliasesFunc | None = None, /) -> FilterAliasesFunc:
        """
        Decorate a function to add additional queryset aliases required by this Filter.

        >>> class TaskFilterSet(FilterSet[Task]):
        ...     name = Filter()
        ...
        ...     @name.aliases
        ...     def name_aliases(self: Filter, info: GQLInfo, *, value: str) -> dict[str, DjangoExpression]:
        ...         return {"foo": Value("bar")}
        """
        if func is None:  # Allow `@<filter_name>.aliases()`
            return self.aliases  # type: ignore[return-value]
        self.aliases_func = get_wrapped_func(func)
        return func

    def visible(self, func: VisibilityFunc | None = None, /) -> VisibilityFunc:
        """
        Decorate a function to change the Filter's visibility in the schema.
        Experimental, requires `EXPERIMENTAL_VISIBILITY_CHECKS` to be enabled.

        >>> class TaskFilterSet(FilterSet[Task]):
        ...     name = Filter()
        ...
        ...     @name.visible
        ...     def name_visible(self: Filter, request: DjangoRequestProtocol) -> bool:
        ...         return False
        """
        if func is None:  # Allow `@<filter_name>.visible()`
            return self.visible  # type: ignore[return-value]
        self.visible_func = get_wrapped_func(func)
        return func

    def add_directive(self, directive: Directive, /) -> Self:
        """Add a directive to this filter."""
        check_directives([directive], location=DirectiveLocation.INPUT_FIELD_DEFINITION)
        self.directives.append(directive)
        return self


def get_filters_for_model(model: type[Model], *, exclude: Iterable[str] = ()) -> dict[str, Filter]:
    """Creates `Filters` for all the given Model's fields, except those in the 'exclude' list."""
    result: dict[str, Filter] = {}

    # Lookups are separated by '__', but auto-generated names use '_' instead.
    exclude = {"_".join(item.split(LOOKUP_SEP)) for item in exclude}

    for model_field in get_model_fields_for_graphql(model):
        field_name = model_field.name

        # Filters for many-to-many relations should always use `qs.distinct()`
        distinct = is_to_many(model_field)

        is_primary_key = bool(getattr(model_field, "primary_key", False))
        if is_primary_key:
            field_name = "pk"

        if field_name in exclude:
            continue

        lookups = sorted(convert_to_filter_lookups(model_field))  # type: ignore[arg-type]

        for lookup in lookups:
            display_name = lookup_to_display_name(lookup, model_field)

            name = f"{field_name}_{display_name}" if display_name else field_name
            if name in exclude:
                continue

            result[name] = Filter(field_name, lookup=lookup, distinct=distinct)

    return result


def get_filters_for_models(models: tuple[type[TModel], ...], *, exclude: Iterable[str] = ()) -> dict[str, Filter]:
    result: dict[str, Filter] = {}

    # Lookups are separated by '__', but auto-generated names use '_' instead.
    exclude = {"_".join(item.split(LOOKUP_SEP)) for item in exclude}

    fields_by_model: dict[type[Model], set[str]] = {}
    for model in models:
        fields: set[str] = set()

        for model_field in get_model_fields_for_graphql(model):
            field_name = model_field.name

            is_primary_key = bool(getattr(model_field, "primary_key", False))
            if is_primary_key:
                field_name = "pk"

            if field_name in exclude:
                continue

            fields.add(field_name)

        fields_by_model[model] = fields

    common_fields = functools.reduce(operator.and_, fields_by_model.values())

    graphql_types_by_model: dict[str, dict[type[Model], GraphQLInputType]] = defaultdict(dict)
    for model in fields_by_model:
        for field_name in common_fields:
            graphql_types_by_model[field_name][model] = convert_to_graphql_type(field_name, model=model, is_input=True)

    usable_fields: set[str] = set()
    for field_name, model_map in graphql_types_by_model.items():
        is_usable = all(field_1 == field_2 for field_1, field_2 in itertools.combinations(model_map.values(), 2))
        if is_usable:
            usable_fields.add(field_name)

    for field_name in usable_fields:
        model_field = get_model_field(model=models[0], lookup=field_name)

        lookups = sorted(convert_to_filter_lookups(model_field))  # type: ignore[arg-type]

        for lookup in lookups:
            # Filters for many-to-many relations should always use `qs.distinct()`
            distinct = is_to_many(model_field)

            display_name = lookup_to_display_name(lookup, model_field)

            name = f"{field_name}_{display_name}" if display_name else field_name
            if name in exclude:
                continue

            result[name] = Filter(field_name, lookup=lookup, distinct=distinct)

    return result



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, Self, Unpack

from graphql import DirectiveLocation, GraphQLField, Undefined

from undine.parsers import parse_class_attribute_docstrings
from undine.query import QueryType
from undine.settings import undine_settings
from undine.utils.graphql.type_registry import get_or_create_graphql_interface_type
from undine.utils.graphql.utils import check_directives
from undine.utils.reflection import FunctionEqualityWrapper, get_members, get_wrapped_func, is_subclass
from undine.utils.text import dotpath, get_docstring, to_schema_name

if TYPE_CHECKING:
    from graphql import GraphQLArgumentMap, GraphQLInterfaceType, GraphQLOutputType

    from undine.directives import Directive
    from undine.query import Field
    from undine.typing import (
        DjangoRequestProtocol,
        InterfaceFieldParams,
        InterfaceTypeParams,
        TInterfaceQueryType,
        VisibilityFunc,
    )

__all__ = [
    "InterfaceField",
    "InterfaceType",
]


class InterfaceTypeMeta(type):
    """A metaclass that modifies how a `InterfaceType` is created."""

    # Set in '__new__'
    __field_map__: dict[str, InterfaceField]
    __schema_name__: str
    __interfaces__: list[type[InterfaceType]]
    __implementations__: list[type[InterfaceType | QueryType]]
    __directives__: list[Directive]
    __extensions__: dict[str, Any]
    __attribute_docstrings__: dict[str, str]

    def __new__(
        cls,
        _name: str,
        _bases: tuple[type, ...],
        _attrs: dict[str, Any],
        **kwargs: Unpack[InterfaceTypeParams],
    ) -> InterfaceTypeMeta:
        if _name == "InterfaceType":  # Early return for the `InterfaceType` class itself.
            return super().__new__(cls, _name, _bases, _attrs)

        interfaces = kwargs.get("interfaces", [])
        interfaces = get_with_inherited_interfaces(interfaces)

        for interface in interfaces:
            for field_name, interface_field in interface.__field_map__.items():
                _attrs.setdefault(field_name, interface_field)

        interface_type = super().__new__(cls, _name, _bases, _attrs)

        # Members should use `__dunder__` names to avoid name collisions with possible `InterfaceField` names.
        interface_type.__field_map__ = get_members(interface_type, InterfaceField)
        interface_type.__schema_name__ = kwargs.get("schema_name", _name)
        interface_type.__interfaces__ = interfaces
        interface_type.__implementations__ = []
        interface_type.__directives__ = kwargs.get("directives", [])
        interface_type.__extensions__ = kwargs.get("extensions", {})
        interface_type.__attribute_docstrings__ = parse_class_attribute_docstrings(interface_type)

        check_directives(interface_type.__directives__, location=DirectiveLocation.INTERFACE)
        interface_type.__extensions__[undine_settings.INTERFACE_TYPE_EXTENSIONS_KEY] = interface_type

        for interface in interfaces:
            interface.__register_as_implementation__(interface_type)  # type: ignore[arg-type]

        for name, interface_field in interface_type.__field_map__.items():
            interface_field.__connect__(interface_type, name)  # type: ignore[arg-type]

        return interface_type

    def __str__(cls) -> str:
        return undine_settings.SDL_PRINTER.print_interface_type(cls.__interface__())

    def __call__(cls, implementation: type[TInterfaceQueryType]) -> type[TInterfaceQueryType]:
        """
        Allow iheriting this InterfaceType to a QueryType or another InterfaceType using a decorator syntax.

        >>> class Named(InterfaceType): ...
        >>>
        >>> @Named
        >>> class TaskType(QueryType[Task]): ...
        """
        if is_subclass(implementation, QueryType):
            for field_name, interface_field in cls.__field_map__.items():
                field = interface_field.as_undine_field()
                setattr(implementation, field_name, field)
                implementation.__field_map__[field_name] = field
                field.__connect__(implementation, field_name)

        elif is_subclass(implementation, InterfaceType):
            for field_name, interface_field in cls.__field_map__.items():
                setattr(implementation, field_name, interface_field)
                implementation.__field_map__[field_name] = interface_field

        implementation.__interfaces__.append(cls)  # type: ignore[assignment]
        cls.__register_as_implementation__(implementation)
        return implementation

    def __register_as_implementation__(cls, implementation: type[InterfaceType | QueryType]) -> None:
        cls.__implementations__.append(implementation)
        for interface in cls.__interfaces__:
            interface.__register_as_implementation__(implementation)

    def __concrete_implementations__(cls) -> list[type[QueryType]]:
        return [impl for impl in cls.__implementations__ if not issubclass(impl, InterfaceType)]  # type: ignore[return-value]

    def __interface__(cls) -> GraphQLInterfaceType:
        return get_or_create_graphql_interface_type(
            name=cls.__schema_name__,
            fields=FunctionEqualityWrapper(cls.__output_fields__, context=cls),
            interfaces=[instance.__interface__() for instance in cls.__interfaces__],
            description=get_docstring(cls),
            extensions=cls.__extensions__,
        )

    def __output_fields__(cls) -> dict[str, GraphQLField]:
        """Defer creating fields until all QueryTypes have been registered."""
        return {field.schema_name: field.as_graphql_field() for field in cls.__field_map__.values()}

    def __add_directive__(cls, directive: Directive, /) -> Self:
        """Add a directive to this interface."""
        check_directives([directive], location=DirectiveLocation.INTERFACE)
        cls.__directives__.append(directive)
        return cls


class InterfaceType(metaclass=InterfaceTypeMeta):
    """
    Class for creating a new `InterfaceType` for a `QueryType`.
    Represents a GraphQL `GraphQLInterfaceType` in the GraphQL schema.

    The following parameters can be passed in the class definition:

     `interfaces: list[type[InterfaceType]] = []`
        Interfaces this `InterfaceType` should implement.

     `schema_name: str = <class name>`
        Override name for the `GraphQLInterfaceType` for this `InterfaceType` in the GraphQL schema.

     `directives: list[Directive] = []`
        `Directives` to add to the created `GraphQLInterfaceType`.

     `extensions: dict[str, Any] = {}`
        GraphQL extensions for the created `GraphQLInterfaceType`.

    >>> class Node(InterfaceType)
    >>>     id = InterfaceField(GraphQLNonNull(GraphQLID), field_name="pk")
    """

    # Set in metaclass
    __field_map__: ClassVar[dict[str, InterfaceField]]
    __schema_name__: ClassVar[str]
    __interfaces__: ClassVar[list[type[InterfaceType]]]
    __implementations__: ClassVar[list[type[InterfaceType | QueryType]]]
    __directives__: ClassVar[list[Directive]]
    __extensions__: ClassVar[dict[str, Any]]
    __attribute_docstrings__: ClassVar[dict[str, str]]

    @classmethod
    def __is_visible__(cls, request: DjangoRequestProtocol) -> bool:
        """
        Determine if the given `InterfaceType` is visible in the schema.
        Experimental, requires `EXPERIMENTAL_VISIBILITY_CHECKS` to be enabled.
        """
        return True


class InterfaceField:
    """
    A class for defining a field for an `InterfaceType`.
    Represents a field on a GraphQL `Interface` for the `InterfaceType` this is added to.

    >>> class Named(InterfaceType):
    ...     name = InterfaceField(GraphQLNonNull(GraphQLString))
    """

    def __init__(self, output_type: GraphQLOutputType, **kwargs: Unpack[InterfaceFieldParams]) -> None:
        """
        Create a new `InterfaceField`.

        :param output_type: The output type to use for the `InterfaceField`.
        :param args: GraphQL arguments for the `InterfaceField`.
        :param description: Description for the `InterfaceField`.
        :param deprecation_reason: If the `InterfaceField` is deprecated, describes the reason for deprecation.
        :param field_name: The name of the field in the Django model. If not provided, use the name of the attribute.
        :param schema_name: Actual name in the GraphQL schema. Only needed if argument name is a python keyword.
        :param directives: GraphQL directives for the `InterfaceField`.
        :param extensions: GraphQL extensions for the `InterfaceField`.
        """
        self.output_type = output_type
        self.args: GraphQLArgumentMap = kwargs.get("args", {})
        self.description: str | None = kwargs.get("description", Undefined)  # type: ignore[assignment]
        self.deprecation_reason: str | None = kwargs.get("deprecation_reason")
        self.field_name: str = kwargs.get("field_name", Undefined)  # type: ignore[assignment]
        self.schema_name: str = kwargs.get("schema_name", Undefined)  # type: ignore[assignment]
        self.directives: list[Directive] = kwargs.get("directives", [])
        self.extensions: dict[str, Any] = kwargs.get("extensions", {})

        check_directives(self.directives, location=DirectiveLocation.FIELD_DEFINITION)
        self.extensions[undine_settings.INTERFACE_FIELD_EXTENSIONS_KEY] = self

        self.visible_func: VisibilityFunc | None = None

    def __connect__(self, interface_type: type[InterfaceType], name: str) -> None:
        """Connect this `InterfaceField` to the given `InterfaceType` using the given name."""
        self.interface_type = interface_type
        self.name = name
        self.field_name = self.field_name or name
        self.schema_name = self.schema_name or to_schema_name(name)

        if self.description is Undefined:
            self.description = interface_type.__attribute_docstrings__.get(name)

    def __repr__(self) -> str:
        return f"<{dotpath(self.__class__)}(ref={self.output_type!r})>"

    def __str__(self) -> str:
        field = self.as_graphql_field()
        return undine_settings.SDL_PRINTER.print_field(self.schema_name, field, indent=False)

    def as_graphql_field(self) -> GraphQLField:
        return GraphQLField(
            type_=self.output_type,
            args=self.args,
            description=self.description,
            deprecation_reason=self.deprecation_reason,
            extensions=self.extensions,
        )

    def as_undine_field(self) -> Field:
        """Convert this `InterfaceField` to a `Field` to be added to a `QueryType`."""
        from undine.query import Field  # noqa: PLC0415

        return Field(
            self,
            deprecation_reason=self.deprecation_reason,
            field_name=self.field_name,
            schema_name=self.schema_name,
            directives=self.directives,
            extensions={undine_settings.INTERFACE_FIELD_EXTENSIONS_KEY: self},
        )

    def visible(self, func: VisibilityFunc | None = None, /) -> VisibilityFunc:
        """
        Decorate a function to change the InterfaceField's visibility in the schema.
        Experimental, requires `EXPERIMENTAL_VISIBILITY_CHECKS` to be enabled.

        >>> class Named(InterfaceType):
        ...     name = InterfaceField(GraphQLNonNull(GraphQLString))
        ...
        ...     @name.visible
        ...     def name_visible(self: InterfaceField, request: DjangoRequestProtocol) -> bool:
        ...         return False
        """
        if func is None:  # Allow `@<interface_field_name>.visible()`
            return self.visible  # type: ignore[return-value]
        self.visible_func = get_wrapped_func(func)
        return func

    def add_directive(self, directive: Directive, /) -> Self:
        """Add a directive to this interface field."""
        check_directives([directive], location=DirectiveLocation.FIELD_DEFINITION)
        self.directives.append(directive)
        return self


def get_with_inherited_interfaces(interfaces: list[type[InterfaceType]]) -> list[type[InterfaceType]]:
    """
    Given the list of interfaces that an `InterfaceType` might explicitly inherit,
    add all implicitly inherited interfaces to the list (e.g. interfaces of interfaces).
    """
    all_interfaces: set[type[InterfaceType]] = set()

    for interface in interfaces:
        if interface in all_interfaces:
            continue

        all_interfaces.add(interface)
        all_interfaces.update(get_with_inherited_interfaces(interface.__interfaces__))

    return list(all_interfaces)



from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING, Any

from django.db.models import Q

from undine.utils.reflection import get_root_and_info_params

if TYPE_CHECKING:
    from collections.abc import Callable
    from types import FunctionType

    from undine.typing import GQLInfo

__all__ = [
    "FilterFunctionResolver",
    "FilterModelFieldResolver",
]


@dataclasses.dataclass(frozen=True, slots=True)
class FilterFunctionResolver:
    """Resolves a `FilterSet` field function."""

    func: FunctionType | Callable[..., Any] = dataclasses.field(kw_only=True)
    root_param: str | None = dataclasses.field(default=None, init=False)
    info_param: str | None = dataclasses.field(default=None, init=False)

    def __post_init__(self) -> None:
        params = get_root_and_info_params(self.func)
        object.__setattr__(self, "root_param", params.root_param)
        object.__setattr__(self, "info_param", params.info_param)

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> Q:
        if self.root_param is not None:
            kwargs[self.root_param] = root
        if self.info_param is not None:
            kwargs[self.info_param] = info
        return self.func(**kwargs)


@dataclasses.dataclass(frozen=True, slots=True)
class FilterModelFieldResolver:
    """Resolves a filter to a model field lookup."""

    lookup: str

    def __call__(self, root: Any, info: GQLInfo, *, value: Any) -> Q:
        return Q(**{self.lookup: value})


@dataclasses.dataclass(frozen=True, slots=True)
class FilterQExpressionResolver:
    """Resolves a filter using a Q expression."""

    q_expression: Q

    def __call__(self, root: Any, info: GQLInfo, *, value: bool) -> Q:
        return self.q_expression if value else ~self.q_expression



from __future__ import annotations

import base64

from ._definition import ScalarType

__all__ = [
    "GraphQLBase32",
    "base32_scalar",
]


base32_scalar: ScalarType[str, str] = ScalarType(
    name="Base32",
    description="Represents a base32-encoded string.",
    specified_by_url="https://datatracker.ietf.org/doc/html/rfc4648#section-6",
)

GraphQLBase32 = base32_scalar.as_graphql_scalar()


@base32_scalar.serialize.register
@base32_scalar.parse.register
def _(value: str) -> str:
    # Validate string is base32 encoded
    base64.b32decode(value.encode(encoding="utf-8")).decode(encoding="utf-8")
    return value


@base32_scalar.serialize.register
def _(value: bytes) -> str:
    # Validate string is base32 encoded
    base64.b32decode(value).decode(encoding="utf-8")
    return value.decode(encoding="utf-8")



from __future__ import annotations

from types import FunctionType
from typing import Any

from graphql import UndefinedType

from undine import Entrypoint, InterfaceType, MutationType, QueryType, UnionType
from undine.converters import convert_to_entrypoint_ref, is_many
from undine.exceptions import MissingEntrypointRefError
from undine.pagination import OffsetPagination
from undine.parsers import parse_is_nullable
from undine.relay import Connection, Node
from undine.settings import undine_settings
from undine.subscriptions import SignalSubscription


@convert_to_entrypoint_ref.register
def _(_: UndefinedType, **kwargs: Any) -> Any:
    caller: Entrypoint = kwargs["caller"]
    raise MissingEntrypointRefError(name=caller.name, cls=caller.root_type)


@convert_to_entrypoint_ref.register
def _(ref: FunctionType, **kwargs: Any) -> Any:
    caller: Entrypoint = kwargs["caller"]
    caller.many = is_many(ref)
    caller.nullable = parse_is_nullable(ref)
    return ref


@convert_to_entrypoint_ref.register
def _(ref: type[QueryType], **kwargs: Any) -> Any:
    return ref


@convert_to_entrypoint_ref.register
def _(ref: type[MutationType], **kwargs: Any) -> Any:
    return ref


@convert_to_entrypoint_ref.register
def _(ref: type[UnionType], **kwargs: Any) -> Any:
    caller: Entrypoint = kwargs["caller"]
    caller.many = True
    return ref


@convert_to_entrypoint_ref.register
def _(ref: type[InterfaceType], **kwargs: Any) -> Any:
    caller: Entrypoint = kwargs["caller"]
    caller.many = True
    return ref


@convert_to_entrypoint_ref.register
def _(ref: type[Node], **kwargs: Any) -> Any:
    caller: Entrypoint = kwargs["caller"]
    caller.many = False
    return ref


@convert_to_entrypoint_ref.register
def _(ref: Connection, **kwargs: Any) -> Any:
    caller: Entrypoint = kwargs["caller"]
    caller.many = False
    caller.extensions[undine_settings.CONNECTION_EXTENSIONS_KEY] = ref
    return ref


@convert_to_entrypoint_ref.register
def _(ref: OffsetPagination, **kwargs: Any) -> Any:
    caller: Entrypoint = kwargs["caller"]
    caller.many = True
    caller.extensions[undine_settings.OFFSET_PAGINATION_EXTENSIONS_KEY] = ref
    return ref


@convert_to_entrypoint_ref.register
def _(ref: SignalSubscription, **kwargs: Any) -> Any:
    caller: Entrypoint = kwargs["caller"]
    caller.many = False
    return ref



from __future__ import annotations

import inspect
from types import FunctionType
from typing import TYPE_CHECKING, Any

from asgiref.sync import sync_to_async
from graphql import GraphQLError, Undefined

from undine.exceptions import GraphQLErrorGroup
from undine.utils.graphql.utils import graphql_error_path
from undine.utils.model_utils import get_instance_or_raise, get_instances_or_raise

if TYPE_CHECKING:
    from django.db.models import Model

    from undine import GQLInfo, MutationType
    from undine.typing import MutationDataCoroutine, MutationDataFunc


__all__ = [
    "pre_mutation",
    "pre_mutation_async",
    "pre_mutation_many",
    "pre_mutation_many_async",
]


# Sync


def pre_mutation(
    instance: Model,
    info: GQLInfo,
    input_data: dict[str, Any],
    mutation_type: type[MutationType],
) -> None:
    """Run all pre-mutation handling for the given instance and input data."""
    _fetch_model_inputs(
        instance=instance,
        info=info,
        input_data=input_data,
        mutation_type=mutation_type,
    )
    _add_hidden_inputs(
        instance=instance,
        info=info,
        input_data=input_data,
        mutation_type=mutation_type,
    )
    _run_function_inputs(
        instance=instance,
        info=info,
        input_data=input_data,
        mutation_type=mutation_type,
    )
    _permissions(
        instance=instance,
        info=info,
        input_data=input_data,
        mutation_type=mutation_type,
    )
    _validate(
        instance=instance,
        info=info,
        input_data=input_data,
        mutation_type=mutation_type,
    )
    _remove_input_only_inputs(
        instance=instance,
        info=info,
        input_data=input_data,
        mutation_type=mutation_type,
    )


def pre_mutation_many(
    instances: list[Model],
    info: GQLInfo,
    input_data: list[dict[str, Any]],
    mutation_type: type[MutationType],
) -> None:
    """Run all pre-mutation handling for the given instances and input data."""
    errors: list[GraphQLError] = []

    for i, (instance, sub_data) in enumerate(zip(instances, input_data, strict=True)):
        try:
            with graphql_error_path(info, key=i) as sub_info:
                pre_mutation(
                    instance=instance,
                    info=sub_info,
                    input_data=sub_data,
                    mutation_type=mutation_type,
                )
        except GraphQLError as error:
            errors.append(error)

        except GraphQLErrorGroup as error_group:
            errors.extend(error_group.flatten())

    if errors:
        raise GraphQLErrorGroup(errors)


def _fetch_model_inputs(
    instance: Model,
    info: GQLInfo,
    input_data: dict[str, Any],
    mutation_type: type[MutationType],
) -> None:
    for input_field in mutation_type.__model_inputs__.values():
        field_data = input_data.get(input_field.name)
        if field_data is None:
            continue

        with graphql_error_path(info, key=input_field.name):
            if input_field.many:
                input_data[input_field.name] = get_instances_or_raise(model=input_field.ref, pks=field_data)
            else:
                input_data[input_field.name] = get_instance_or_raise(model=input_field.ref, pk=field_data)

    _run_for_related_mutation_types(
        func=_fetch_model_inputs,
        mutation_type=mutation_type,
        info=info,
        input_data=input_data,
    )


def _add_hidden_inputs(
    instance: Model,
    info: GQLInfo,
    input_data: dict[str, Any],
    mutation_type: type[MutationType],
) -> None:
    for input_field in mutation_type.__hidden_inputs__.values():
        if isinstance(input_field.ref, FunctionType):
            input_data[input_field.name] = input_field.ref(instance, info)

        elif input_field.default_value is not Undefined:
            input_data[input_field.name] = input_field.default_value

    _run_for_related_mutation_types(
        func=_add_hidden_inputs,
        mutation_type=mutation_type,
        info=info,
        input_data=input_data,
    )


def _run_function_inputs(
    instance: Model,
    info: GQLInfo,
    input_data: dict[str, Any],
    mutation_type: type[MutationType],
) -> None:
    for input_field in mutation_type.__function_inputs__.values():
        if input_field.hidden:  # Hidden function inputs have already been run
            continue

        field_data = input_data.get(input_field.name, Undefined)
        if field_data is Undefined:
            continue

        input_data[input_field.name] = input_field.ref(instance, info, field_data)

    _run_for_related_mutation_types(
        func=_run_function_inputs,
        mutation_type=mutation_type,
        info=info,
        input_data=input_data,
    )


def _permissions(
    instance: Model,
    info: GQLInfo,
    input_data: dict[str, Any],
    mutation_type: type[MutationType],
) -> None:
    with graphql_error_path(info):
        mutation_type.__permissions__(instance, info, input_data)

    errors: list[GraphQLError] = []

    for key, value in input_data.items():
        input_field = mutation_type.__input_map__[key]
        if input_field.permissions_func is None:
            continue

        if value == input_field.default_value:
            continue

        try:
            with graphql_error_path(info, key=key) as sub_info:
                input_field.permissions_func(instance, sub_info, value)

        except GraphQLError as error:
            errors.append(error)

        except GraphQLErrorGroup as error_group:
            errors.extend(error_group.flatten())

    if errors:
        raise GraphQLErrorGroup(errors)

    _run_for_related_mutation_types(
        func=_permissions,
        mutation_type=mutation_type,
        info=info,
        input_data=input_data,
    )


def _validate(
    instance: Model,
    info: GQLInfo,
    input_data: dict[str, Any],
    mutation_type: type[MutationType],
) -> None:
    errors: list[GraphQLError] = []

    for key, value in input_data.items():
        input_field = mutation_type.__input_map__[key]
        if input_field.validator_func is None:
            continue

        if value == input_field.default_value:
            continue

        try:
            with graphql_error_path(info, key=key) as sub_info:
                input_field.validator_func(instance, sub_info, value)

        except GraphQLError as error:
            errors.append(error)

        except GraphQLErrorGroup as error_group:
            errors.extend(error_group.flatten())

    if errors:
        raise GraphQLErrorGroup(errors)

    _run_for_related_mutation_types(
        func=_validate,
        mutation_type=mutation_type,
        info=info,
        input_data=input_data,
    )

    with graphql_error_path(info):
        mutation_type.__validate__(instance, info, input_data)


def _remove_input_only_inputs(
    instance: Model,
    info: GQLInfo,
    input_data: dict[str, Any],
    mutation_type: type[MutationType],
) -> None:
    for input_field in mutation_type.__input_only_inputs__.values():
        input_data.pop(input_field.name, None)

    _run_for_related_mutation_types(
        func=_remove_input_only_inputs,
        mutation_type=mutation_type,
        info=info,
        input_data=input_data,
    )


def _run_for_related_mutation_types(
    func: MutationDataFunc,
    mutation_type: type[MutationType],
    info: GQLInfo,
    input_data: dict[str, Any],
) -> None:
    errors: list[GraphQLError] = []

    for input_field in mutation_type.__related_inputs__.values():
        field_data: dict[str, Any] | list[dict[str, Any]] | None = input_data.get(input_field.name)
        if not field_data:  # No value, null, empty list, empty dict, etc.
            continue

        related_mutation_type: type[MutationType] = input_field.ref
        model = related_mutation_type.__model__

        with graphql_error_path(info, key=input_field.name) as sub_info:
            if not input_field.many:
                dict_data: dict[str, Any] = field_data  # type: ignore[assignment]

                try:
                    func(
                        instance=model(),
                        info=sub_info,
                        input_data=dict_data,
                        mutation_type=related_mutation_type,
                    )

                except GraphQLError as error:
                    errors.append(error)

                except GraphQLErrorGroup as error_group:
                    errors.extend(error_group.flatten())

                continue

            list_data: list[dict[str, Any]] = field_data  # type: ignore[assignment]
            for i, item in enumerate(list_data):
                try:
                    with graphql_error_path(sub_info, key=i) as list_info:
                        func(
                            instance=model(),
                            info=list_info,
                            input_data=item,
                            mutation_type=related_mutation_type,
                        )

                except GraphQLError as error:
                    errors.append(error)

                except GraphQLErrorGroup as error_group:
                    errors.extend(error_group.flatten())
    if errors:
        raise GraphQLErrorGroup(errors)


# Async


async def pre_mutation_async(
    instance: Model,
    info: GQLInfo,
    input_data: dict[str, Any],
    mutation_type: type[MutationType],
) -> None:
    """Run all pre-mutation handling for the given instance and input data."""
    await _fetch_model_inputs_async(
        instance=instance,
        info=info,
        input_data=input_data,
        mutation_type=mutation_type,
    )
    await _add_hidden_inputs_async(
        instance=instance,
        info=info,
        input_data=input_data,
        mutation_type=mutation_type,
    )
    await _run_function_inputs_async(
        instance=instance,
        info=info,
        input_data=input_data,
        mutation_type=mutation_type,
    )
    await _permissions_async(
        instance=instance,
        info=info,
        input_data=input_data,
        mutation_type=mutation_type,
    )
    await _validate_async(
        instance=instance,
        info=info,
        input_data=input_data,
        mutation_type=mutation_type,
    )
    _remove_input_only_inputs(
        instance=instance,
        info=info,
        input_data=input_data,
        mutation_type=mutation_type,
    )


async def pre_mutation_many_async(
    instances: list[Model],
    info: GQLInfo,
    input_data: list[dict[str, Any]],
    mutation_type: type[MutationType],
) -> None:
    """Run all pre-mutation handling for the given instances and input data."""
    errors: list[GraphQLError] = []

    for i, (instance, sub_data) in enumerate(zip(instances, input_data, strict=True)):
        try:
            with graphql_error_path(info, key=i) as sub_info:
                await pre_mutation_async(
                    instance=instance,
                    info=sub_info,
                    input_data=sub_data,
                    mutation_type=mutation_type,
                )
        except GraphQLError as error:
            errors.append(error)

        except GraphQLErrorGroup as error_group:
            errors.extend(error_group.flatten())

    if errors:
        raise GraphQLErrorGroup(errors)


async def _fetch_model_inputs_async(
    instance: Model,
    info: GQLInfo,
    input_data: dict[str, Any],
    mutation_type: type[MutationType],
) -> None:
    for input_field in mutation_type.__model_inputs__.values():
        field_data = input_data.get(input_field.name)
        if field_data is None:
            continue

        with graphql_error_path(info, key=input_field.name):
            if input_field.many:
                input_data[input_field.name] = await sync_to_async(get_instances_or_raise)(
                    model=input_field.ref,
                    pks=field_data,
                )
            else:
                input_data[input_field.name] = await sync_to_async(get_instance_or_raise)(
                    model=input_field.ref,
                    pk=field_data,
                )

    await _run_for_related_mutation_types_async(
        coroutine=_fetch_model_inputs_async,
        mutation_type=mutation_type,
        info=info,
        input_data=input_data,
    )


async def _add_hidden_inputs_async(
    instance: Model,
    info: GQLInfo,
    input_data: dict[str, Any],
    mutation_type: type[MutationType],
) -> None:
    for input_field in mutation_type.__hidden_inputs__.values():
        if inspect.iscoroutinefunction(input_field.ref):
            input_data[input_field.name] = await input_field.ref(instance, info)

        elif isinstance(input_field.ref, FunctionType):
            input_data[input_field.name] = input_field.ref(instance, info)

        elif input_field.default_value is not Undefined:
            input_data[input_field.name] = input_field.default_value

    await _run_for_related_mutation_types_async(
        coroutine=_add_hidden_inputs_async,
        mutation_type=mutation_type,
        info=info,
        input_data=input_data,
    )


async def _run_function_inputs_async(
    instance: Model,
    info: GQLInfo,
    input_data: dict[str, Any],
    mutation_type: type[MutationType],
) -> None:
    for input_field in mutation_type.__function_inputs__.values():
        if input_field.hidden:  # Hidden function inputs have already been run
            continue

        field_data = input_data.get(input_field.name, Undefined)
        if field_data is Undefined:
            continue

        if inspect.iscoroutinefunction(input_field.ref):
            input_data[input_field.name] = await input_field.ref(instance, info, field_data)
            continue

        input_data[input_field.name] = input_field.ref(instance, info, field_data)

    await _run_for_related_mutation_types_async(
        coroutine=_run_function_inputs_async,
        mutation_type=mutation_type,
        info=info,
        input_data=input_data,
    )


async def _permissions_async(
    instance: Model,
    info: GQLInfo,
    input_data: dict[str, Any],
    mutation_type: type[MutationType],
) -> None:
    with graphql_error_path(info):
        if inspect.iscoroutinefunction(mutation_type.__permissions__):
            await mutation_type.__permissions__(instance, info, input_data)
        else:
            mutation_type.__permissions__(instance, info, input_data)

    errors: list[GraphQLError] = []

    for key, value in input_data.items():
        input_field = mutation_type.__input_map__[key]
        if input_field.permissions_func is None:
            continue

        if value == input_field.default_value:
            continue

        try:
            with graphql_error_path(info, key=key) as sub_info:
                if inspect.iscoroutinefunction(input_field.permissions_func):
                    await input_field.permissions_func(instance, sub_info, value)
                else:
                    input_field.permissions_func(instance, sub_info, value)

        except GraphQLError as error:
            errors.append(error)

        except GraphQLErrorGroup as error_group:
            errors.extend(error_group.flatten())

    if errors:
        raise GraphQLErrorGroup(errors)

    await _run_for_related_mutation_types_async(
        coroutine=_permissions_async,
        mutation_type=mutation_type,
        info=info,
        input_data=input_data,
    )


async def _validate_async(
    instance: Model,
    info: GQLInfo,
    input_data: dict[str, Any],
    mutation_type: type[MutationType],
) -> None:
    errors: list[GraphQLError] = []

    for key, value in input_data.items():
        input_field = mutation_type.__input_map__[key]
        if input_field.validator_func is None:
            continue

        if value == input_field.default_value:
            continue

        try:
            with graphql_error_path(info, key=key) as sub_info:
                if inspect.iscoroutinefunction(input_field.validator_func):
                    await input_field.validator_func(instance, sub_info, value)
                else:
                    input_field.validator_func(instance, sub_info, value)

        except GraphQLError as error:
            errors.append(error)

        except GraphQLErrorGroup as error_group:
            errors.extend(error_group.flatten())

    if errors:
        raise GraphQLErrorGroup(errors)

    await _run_for_related_mutation_types_async(
        coroutine=_validate_async,
        mutation_type=mutation_type,
        info=info,
        input_data=input_data,
    )

    with graphql_error_path(info):
        if inspect.iscoroutinefunction(mutation_type.__validate__):
            await mutation_type.__validate__(instance, info, input_data)
        else:
            mutation_type.__validate__(instance, info, input_data)


async def _run_for_related_mutation_types_async(
    coroutine: MutationDataCoroutine,
    mutation_type: type[MutationType],
    info: GQLInfo,
    input_data: dict[str, Any],
) -> None:
    errors: list[GraphQLError] = []

    for input_field in mutation_type.__related_inputs__.values():
        field_data: dict[str, Any] | list[dict[str, Any]] | None = input_data.get(input_field.name)
        if not field_data:  # No value, null, empty list, empty dict, etc.
            continue

        related_mutation_type: type[MutationType] = input_field.ref
        model = related_mutation_type.__model__

        with graphql_error_path(info, key=input_field.name) as sub_info:
            if not input_field.many:
                dict_data: dict[str, Any] = field_data  # type: ignore[assignment]

                try:
                    await coroutine(
                        instance=model(),
                        info=sub_info,
                        input_data=dict_data,
                        mutation_type=related_mutation_type,
                    )

                except GraphQLError as error:
                    errors.append(error)

                except GraphQLErrorGroup as error_group:
                    errors.extend(error_group.flatten())

                continue

            list_data: list[dict[str, Any]] = field_data  # type: ignore[assignment]
            for i, item in enumerate(list_data):
                try:
                    with graphql_error_path(sub_info, key=i) as list_info:
                        await coroutine(
                            instance=model(),
                            info=list_info,
                            input_data=item,
                            mutation_type=related_mutation_type,
                        )

                except GraphQLError as error:
                    errors.append(error)

                except GraphQLErrorGroup as error_group:
                    errors.extend(error_group.flatten())

    if errors:
        raise GraphQLErrorGroup(errors)



from __future__ import annotations

from contextlib import suppress
from typing import Any

from django.contrib.contenttypes.fields import GenericForeignKey
from django.db.models import F, Model, Q
from graphql import GraphQLNamedType, GraphQLWrappingType

from undine import InterfaceField
from undine.converters import convert_to_description
from undine.dataclasses import LazyGenericForeignKey, LazyLambda, LazyRelation, TypeRef
from undine.pagination import OffsetPagination
from undine.parsers import docstring_parser
from undine.relay import Connection
from undine.subscriptions import SignalSubscription
from undine.typing import CombinableExpression, ModelField
from undine.utils.text import get_docstring


@convert_to_description.register
def _(ref: Any, **kwargs: Any) -> Any:
    docstring = get_docstring(ref)
    return docstring_parser.parse_body(docstring)


@convert_to_description.register
def _(ref: ModelField, **kwargs: Any) -> Any:
    help_text = getattr(ref, "help_text", None) or None
    if help_text is None:
        return None
    return str(help_text)


@convert_to_description.register
def _(_: type[Model], **kwargs: Any) -> Any:
    return None


@convert_to_description.register
def _(_: CombinableExpression | F | Q, **kwargs: Any) -> Any:
    return None


@convert_to_description.register
def _(_: TypeRef, **kwargs: Any) -> Any:
    return None


@convert_to_description.register
def _(_: type[tuple], **kwargs: Any) -> Any:
    # For NamedTuples, don't use the generated docstring.
    return None


@convert_to_description.register
def _(ref: LazyRelation, **kwargs: Any) -> Any:
    return convert_to_description(ref.field)


@convert_to_description.register
def _(ref: LazyGenericForeignKey, **kwargs: Any) -> Any:
    return convert_to_description(ref.field)


@convert_to_description.register
def _(_: LazyLambda, **kwargs: Any) -> Any:
    return None


@convert_to_description.register
def _(_: GraphQLNamedType, **kwargs: Any) -> Any:
    return None


@convert_to_description.register
def _(ref: GraphQLWrappingType, **kwargs: Any) -> Any:
    return convert_to_description(ref.of_type)


@convert_to_description.register
def _(ref: GenericForeignKey, **kwargs: Any) -> Any:  # Required for Django<5.1
    return getattr(ref, "help_text", None) or None


@convert_to_description.register
def _(ref: Connection, **kwargs: Any) -> Any:
    return ref.description


@convert_to_description.register
def _(ref: OffsetPagination, **kwargs: Any) -> Any:
    if ref.description is not None:
        return ref.description
    if ref.interface_type:
        return convert_to_description(ref.interface_type, **kwargs)
    if ref.union_type:
        return convert_to_description(ref.union_type, **kwargs)
    return convert_to_description(ref.query_type, **kwargs)


@convert_to_description.register
def _(ref: InterfaceField, **kwargs: Any) -> Any:
    return ref.description


@convert_to_description.register
def _(ref: SignalSubscription, **kwargs: Any) -> Any:
    return ref.description


with suppress(ImportError):
    from undine.utils.full_text_search import PostgresFTS

    @convert_to_description.register
    def _(_: PostgresFTS, **kwargs: Any) -> Any:
        return None



from __future__ import annotations

from django.db.models import CharField, DateTimeField, Model, TextField

from .validators import validate_document, validate_document_id

__all__ = [
    "PersistedDocument",
    "PersistedDocumentBase",
]


class PersistedDocumentBase(Model):
    """Persisted document base class."""

    document_id = CharField(max_length=255, primary_key=True, validators=[validate_document_id])
    document = TextField(validators=[validate_document])
    created_at = DateTimeField(auto_now_add=True)

    class Meta:
        abstract = True
        verbose_name = "persisted document"
        verbose_name_plural = "persisted documents"

    def __str__(self) -> str:
        return f"Persisted document '{self.document_id}'"


class PersistedDocument(PersistedDocumentBase):
    """Persisted document."""

    class Meta(PersistedDocumentBase.Meta):
        swappable = "UNDINE_PERSISTED_DOCUMENTS_MODEL"



from __future__ import annotations

from copy import deepcopy
from typing import Any

from django.db.models import Expression, F, OuterRef, Q, Subquery
from django.db.models.constants import LOOKUP_SEP
from django.db.models.expressions import ResolvedOuterRef

from undine.converters import extend_expression


@extend_expression.register
def _(expression: F, **kwargs: Any) -> F:
    field_name: str = kwargs["field_name"]
    expression = deepcopy(expression)
    expression.name = f"{field_name}{LOOKUP_SEP}{expression.name}"
    return expression


@extend_expression.register
def _(expression: Q, **kwargs: Any) -> Q:
    field_name: str = kwargs["field_name"]
    expression = deepcopy(expression)

    children = expression.children
    expression.children = []
    for child in children:
        if isinstance(child, tuple):
            key = f"{field_name}{LOOKUP_SEP}{child[0]}"
            value = child[1]

            if value in extend_expression:
                value = extend_expression(child[1], field_name=field_name)

            expression.children.append((key, value))
            continue

        new_expression = extend_expression(child, field_name=field_name)
        expression.children.append(new_expression)

    return expression


@extend_expression.register
def _(expression: Expression, **kwargs: Any) -> Expression:
    field_name: str = kwargs["field_name"]
    expression = deepcopy(expression)

    expressions = [extend_expression(expr, field_name=field_name) for expr in expression.get_source_expressions()]
    expression.set_source_expressions(expressions)
    return expression


@extend_expression.register
def _(expression: Subquery, **kwargs: Any) -> Subquery:
    def extend_subquery(expr: Any, *, field_name_: str) -> Any:
        """For sub-queries, only OuterRefs are rewritten."""
        if isinstance(expr, OuterRef | ResolvedOuterRef):
            expr = deepcopy(expr)
            expr.name = f"{field_name_}{LOOKUP_SEP}{expr.name}"
            return expr

        expr = deepcopy(expr)
        expressions = [extend_subquery(expr, field_name_=field_name_) for expr in expr.get_source_expressions()]
        expr.set_source_expressions(expressions)
        return expr

    field_name: str = kwargs["field_name"]
    expression = deepcopy(expression)
    sub_expressions = expression.query.where.children
    expression.query.where.children = []
    for child in sub_expressions:
        expression.query.where.children.append(extend_subquery(child, field_name_=field_name))
    return expression



from __future__ import annotations

import datetime
from inspect import cleandoc

from django.utils import dateparse

from undine.exceptions import GraphQLScalarInvalidValueError

from ._definition import ScalarType

__all__ = [
    "GraphQLDate",
    "date_scalar",
]


date_scalar: ScalarType[datetime.date, str] = ScalarType(
    name="Date",
    description=cleandoc(
        """
        Represents a date value as specified by ISO 8601.
        Maps to the Python `datetime.date` type.
        """
    ),
    specified_by_url="https://datatracker.ietf.org/doc/html/rfc3339#section-5.6",
)

GraphQLDate = date_scalar.as_graphql_scalar()


@date_scalar.parse.register
def _(value: str) -> datetime.date:
    parsed_value: datetime.date | None = dateparse.parse_date(value)
    if parsed_value is None:
        raise GraphQLScalarInvalidValueError(typename="date")

    return parsed_value


@date_scalar.serialize.register
def _(value: str) -> str:
    parsed_value: datetime.date | None = dateparse.parse_date(value)
    if parsed_value is None:
        raise GraphQLScalarInvalidValueError(typename="date")

    return parsed_value.isoformat()


@date_scalar.serialize.register
def _(value: datetime.date) -> str:
    return value.isoformat()


@date_scalar.serialize.register
def _(value: datetime.datetime) -> str:
    return value.date().isoformat()



from __future__ import annotations

import itertools
from contextlib import suppress
from types import FunctionType
from typing import Any

from django.contrib.contenttypes.fields import GenericForeignKey, GenericRel
from django.db.models import F, Model, Q
from django.db.models.fields.related_descriptors import (
    ForwardManyToOneDescriptor,
    ManyToManyDescriptor,
    ReverseManyToOneDescriptor,
    ReverseOneToOneDescriptor,
)
from django.db.models.query_utils import DeferredAttribute
from graphql import GraphQLInputType

from undine import Filter
from undine.converters import convert_to_filter_ref, convert_to_graphql_type
from undine.dataclasses import UnionFilterRef
from undine.exceptions import UnionModelFieldDirectUsageError, UnionModelFieldMismatchError
from undine.typing import CombinableExpression, DjangoExpression, GQLInfo, ModelField
from undine.utils.model_utils import determine_output_field, get_model_field


@convert_to_filter_ref.register
def _(ref: str, **kwargs: Any) -> Any:
    caller: Filter = kwargs["caller"]

    models = caller.filterset.__models__
    if len(models) == 1:
        field = get_model_field(model=models[0], lookup=ref)
        return convert_to_filter_ref(field, **kwargs)

    fields_by_model: dict[type[Model], GraphQLInputType] = {}
    for model in caller.filterset.__models__:
        field = get_model_field(model=model, lookup=caller.field_name)
        fields_by_model[model] = convert_to_graphql_type(field, model=model, is_input=True)

    for (model_1, field_1), (model_2, field_2) in itertools.combinations(fields_by_model.items(), 2):
        if field_1 != field_2:
            raise UnionModelFieldMismatchError(
                ref=ref,
                type_1=field_1,
                model_1=model_1,
                type_2=field_2,
                model_2=model_2,
                kind="Filter",
            )

    return UnionFilterRef(ref=ref, models=tuple(models))


@convert_to_filter_ref.register
def _(_: None, **kwargs: Any) -> Any:
    caller: Filter = kwargs["caller"]
    return convert_to_filter_ref(caller.field_name, **kwargs)


@convert_to_filter_ref.register
def _(_: F, **kwargs: Any) -> Any:
    caller: Filter = kwargs["caller"]
    return convert_to_filter_ref(caller.field_name, **kwargs)


@convert_to_filter_ref.register
def _(ref: ModelField, **kwargs: Any) -> Any:
    caller: Filter = kwargs["caller"]

    models = caller.filterset.__models__
    if len(models) != 1:
        raise UnionModelFieldDirectUsageError(kind="FilterSet")

    return ref


@convert_to_filter_ref.register
def _(ref: DeferredAttribute | ForwardManyToOneDescriptor, **kwargs: Any) -> Any:
    return convert_to_filter_ref(ref.field, **kwargs)


@convert_to_filter_ref.register
def _(ref: ReverseManyToOneDescriptor, **kwargs: Any) -> Any:
    return convert_to_filter_ref(ref.rel, **kwargs)


@convert_to_filter_ref.register
def _(ref: ReverseOneToOneDescriptor, **kwargs: Any) -> Any:
    return convert_to_filter_ref(ref.related, **kwargs)


@convert_to_filter_ref.register
def _(ref: ManyToManyDescriptor, **kwargs: Any) -> Any:
    return convert_to_filter_ref(ref.rel if ref.reverse else ref.field, **kwargs)


@convert_to_filter_ref.register
def _(ref: FunctionType, **kwargs: Any) -> Any:
    return ref


@convert_to_filter_ref.register
def _(ref: Q, **kwargs: Any) -> Any:
    caller: Filter = kwargs["caller"]

    user_func = caller.aliases_func

    def aliases(root: Filter, info: GQLInfo, *, value: bool) -> dict[str, DjangoExpression]:
        results: dict[str, DjangoExpression] = {}
        if user_func is not None:
            results |= user_func(root, info, value=value)

        results[root.name] = ref
        return results

    caller.aliases_func = aliases
    return ref


@convert_to_filter_ref.register
def _(ref: CombinableExpression, **kwargs: Any) -> Any:
    caller: Filter = kwargs["caller"]

    user_func = caller.aliases_func

    def aliases(root: Filter, info: GQLInfo, *, value: Any) -> dict[str, DjangoExpression]:
        results: dict[str, DjangoExpression] = {}
        if user_func is not None:
            results |= user_func(root, info, value=value)

        results[root.name] = ref
        return results

    caller.aliases_func = aliases

    models = caller.filterset.__models__
    if len(models) == 1:
        ref.output_field = determine_output_field(ref, model=models[0])
        return ref

    fields_by_model: dict[type[Model], ModelField] = {}
    for model in models:
        fields_by_model[model] = determine_output_field(ref, model=model)

    for (model_1, field_1), (model_2, field_2) in itertools.combinations(fields_by_model.items(), 2):
        if field_1.__class__ is not field_2.__class__:
            raise UnionModelFieldMismatchError(
                ref=ref,
                type_1=field_1,
                model_1=model_1,
                type_2=field_2,
                model_2=model_2,
                kind="Filter",
            )

    ref.output_field = next(iter(fields_by_model.values()))
    return ref


@convert_to_filter_ref.register
def _(ref: GenericRel, **kwargs: Any) -> Any:
    return convert_to_filter_ref(ref.field, **kwargs)


@convert_to_filter_ref.register  # Required for Django<5.1
def _(ref: GenericForeignKey, **kwargs: Any) -> Any:
    caller: Filter = kwargs["caller"]

    models = caller.filterset.__models__
    if len(models) != 1:
        raise UnionModelFieldDirectUsageError(kind="FilterSet")

    return ref


with suppress(ImportError):
    from undine.utils.full_text_search import PostgresFTS

    @convert_to_filter_ref.register
    def _(ref: PostgresFTS, **kwargs: Any) -> Any:
        caller: Filter = kwargs["caller"]

        user_func = caller.aliases_func

        def aliases(root: Filter, info: GQLInfo, *, value: Any) -> dict[str, DjangoExpression]:
            results: dict[str, DjangoExpression] = {}
            if user_func is not None:
                results |= user_func(root, info, value=value)

            lang = ref.get_search_language(info)
            key = ref.get_vector_alias_key(root, lang)
            results[key] = ref.vectors[lang.name]
            return results

        caller.aliases_func = aliases
        return ref



from __future__ import annotations

import datetime
from inspect import cleandoc

from undine.exceptions import GraphQLScalarInvalidValueError

from ._definition import ScalarType

__all__ = [
    "GraphQLDuration",
    "duration_scalar",
]

duration_scalar: ScalarType[datetime.timedelta, int] = ScalarType(
    name="Duration",
    description=cleandoc(
        """
        Represents a duration of time in seconds.
        Maps to the Python `datetime.timedelta` type.
        """
    ),
)

GraphQLDuration = duration_scalar.as_graphql_scalar()


@duration_scalar.parse.register
def _(value: int) -> datetime.timedelta:
    return datetime.timedelta(seconds=value)


@duration_scalar.parse.register
def _(value: str) -> datetime.timedelta:
    try:
        parsed_value = int(value)
    except ValueError as error:
        raise GraphQLScalarInvalidValueError(typename="integer") from error

    return datetime.timedelta(seconds=parsed_value)


@duration_scalar.serialize.register
def _(value: str) -> int:
    try:
        parsed_value = int(value)
    except ValueError as error:
        raise GraphQLScalarInvalidValueError(typename="integer") from error

    return int(datetime.timedelta(seconds=parsed_value).total_seconds())


@duration_scalar.serialize.register
def _(value: int) -> int:
    return int(datetime.timedelta(seconds=value).total_seconds())


@duration_scalar.serialize.register
def _(value: datetime.timedelta) -> int:
    return int(value.total_seconds())



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, Self, Unpack

from graphql import DirectiveLocation, GraphQLField, Undefined

from undine.converters import (
    convert_to_description,
    convert_to_entrypoint_ref,
    convert_to_entrypoint_resolver,
    convert_to_entrypoint_subscription,
    convert_to_graphql_argument_map,
    convert_to_graphql_type,
)
from undine.dataclasses import MaybeManyOrNonNull
from undine.parsers import parse_class_attribute_docstrings
from undine.settings import undine_settings
from undine.utils.graphql.type_registry import get_or_create_graphql_object_type
from undine.utils.graphql.utils import check_directives
from undine.utils.reflection import FunctionEqualityWrapper, cache_signature_if_function, get_members, get_wrapped_func
from undine.utils.text import dotpath, get_docstring, to_schema_name

if TYPE_CHECKING:
    from graphql import GraphQLArgumentMap, GraphQLFieldResolver, GraphQLObjectType, GraphQLOutputType

    from undine.directives import Directive
    from undine.typing import EntrypointParams, EntrypointPermFunc, RootTypeParams, VisibilityFunc

__all__ = [
    "Entrypoint",
    "RootType",
]


class RootTypeMeta(type):
    """A metaclass that modifies how a `RootType` is created."""

    # Set in '__new__'
    __entrypoint_map__: dict[str, Entrypoint]
    __schema_name__: str
    __directives__: list[Directive]
    __extensions__: dict[str, Any]
    __attribute_docstrings__: dict[str, str]

    def __new__(
        cls,
        _name: str,
        _bases: tuple[type, ...],
        _attrs: dict[str, Any],
        **kwargs: Unpack[RootTypeParams],
    ) -> RootTypeMeta:
        if _name == "RootType":  # Early return for the `RootType` class itself.
            return super().__new__(cls, _name, _bases, _attrs)

        root_type = super().__new__(cls, _name, _bases, _attrs)

        # Members should use `__dunder__` names to avoid name collisions with possible `Entrypoint` names.
        root_type.__entrypoint_map__ = get_members(root_type, Entrypoint)
        root_type.__schema_name__ = kwargs.get("schema_name", _name)
        root_type.__directives__ = kwargs.get("directives", [])
        root_type.__extensions__ = kwargs.get("extensions", {})
        root_type.__attribute_docstrings__ = parse_class_attribute_docstrings(root_type)

        check_directives(root_type.__directives__, location=DirectiveLocation.OBJECT)
        root_type.__extensions__[undine_settings.ROOT_TYPE_EXTENSIONS_KEY] = root_type

        for name, entrypoint in root_type.__entrypoint_map__.items():
            entrypoint.__connect__(root_type, name)  # type: ignore[arg-type]

        return root_type

    def __str__(cls) -> str:
        return undine_settings.SDL_PRINTER.print_object_type(cls.__output_type__())

    def __output_type__(cls) -> GraphQLObjectType:
        """Creates the GraphQL `ObjectType` for this `RootType`."""
        return get_or_create_graphql_object_type(
            name=cls.__schema_name__,
            fields=FunctionEqualityWrapper(cls.__output_fields__, context=cls),
            description=get_docstring(cls),
            extensions=cls.__extensions__,
        )

    def __output_fields__(cls) -> dict[str, GraphQLField]:
        return {
            entrypoint.schema_name: entrypoint.as_graphql_field()  # ...
            for entrypoint in cls.__entrypoint_map__.values()
        }

    def __add_directive__(cls, directive: Directive, /) -> Self:
        """Add a directive to this `RootType`."""
        check_directives([directive], location=DirectiveLocation.OBJECT)
        cls.__directives__.append(directive)
        return cls


class RootType(metaclass=RootTypeMeta):
    """
    A class for creating a new `RootType` with `Entrypoints`.
    Represents a GraphQL `GraphQLObjectType` at the root of the GraphQL schema.

    The following parameters can be passed in the class definition:

    `schema_name: str = <class name>`
        Override name for the `ObjectType` for this `RootType` in the GraphQL schema.

    `directives: list[Directive] = []`
        `Directives` to add to the created `ObjectType`.

    `extensions: dict[str, Any] = {}`
        GraphQL extensions for the created `ObjectType`.

    >>> class TaskType(QueryType[Task]): ...
    >>>
    >>> class Query(RootType):
    ...     tasks = Entrypoint(TaskType, many=True)
    """

    # Members should use `__dunder__` names to avoid name collisions with possible `Entrypoint` names.

    # Set in metaclass
    __entrypoint_map__: ClassVar[dict[str, Entrypoint]]
    __schema_name__: ClassVar[str]
    __directives__: ClassVar[list[Directive]]
    __extensions__: ClassVar[dict[str, Any]]
    __attribute_docstrings__: ClassVar[dict[str, str]]


class Entrypoint:
    """
    A class for creating new fields in the `RootTypes` of the GraphQL schema.
    These are the "entry points" at the top level of the GraphQL schema.

    >>> class TaskType(QueryType[Task]): ...
    >>>
    >>> class Query(RootType):
    ...     tasks = Entrypoint(TaskType, many=True)
    """

    def __init__(self, ref: Any = Undefined, **kwargs: Unpack[EntrypointParams]) -> None:
        """
        Create a new `Entrypoint`.

        :param ref: The reference to use for the `Entrypoint`.
        :param many: Whether the `Entrypoint` should return a non-null list of the referenced type.
        :param nullable: Whether the referenced type can be null.
        :param limit: For list Entrypoints, limits the number of objects that are fetched.
        :param description: Description for the `Entrypoint`.
        :param deprecation_reason: If the `Entrypoint` is deprecated, describes the reason for deprecation.
        :param complexity: The complexity of resolving this field (not the entire Entrypoint).
        :param schema_name: Actual name in the GraphQL schema. Only needed if argument name is a python keyword.
        :param directives: GraphQL directives for the `Entrypoint`.
        :param extensions: GraphQL extensions for the `Entrypoint`.
        """
        self.ref: Any = cache_signature_if_function(ref, depth=1)

        self.many: bool = kwargs.get("many", False)
        self.nullable: bool = kwargs.get("nullable", False)
        self.limit: int | None = kwargs.get("limit", undine_settings.LIST_ENTRYPOINT_LIMIT)
        self.description: str | None = kwargs.get("description", Undefined)  # type: ignore[assignment]
        self.deprecation_reason: str | None = kwargs.get("deprecation_reason")
        self.complexity: int = kwargs.get("complexity", 0)  # type: ignore[assignment]
        self.schema_name: str = kwargs.get("schema_name", Undefined)  # type: ignore[assignment]
        self.directives: list[Directive] = kwargs.get("directives", [])
        self.extensions: dict[str, Any] = kwargs.get("extensions", {})

        check_directives(self.directives, location=DirectiveLocation.FIELD_DEFINITION)
        self.extensions[undine_settings.ENTRYPOINT_EXTENSIONS_KEY] = self

        self.resolver_func: GraphQLFieldResolver | None = None
        self.permissions_func: EntrypointPermFunc | None = None
        self.visible_func: VisibilityFunc | None = None

    def __connect__(self, root_type: type[RootType], name: str) -> None:
        """Connect this `Entrypoint` to the given `RootType` using the given name."""
        self.root_type = root_type
        self.name = name
        self.schema_name = self.schema_name or to_schema_name(name)

        self.ref = convert_to_entrypoint_ref(self.ref, caller=self)

        if self.description is Undefined:
            self.description = self.root_type.__attribute_docstrings__.get(name)
            if self.description is None:
                self.description = convert_to_description(self.ref)

    def __call__(self, ref: GraphQLFieldResolver, /) -> Entrypoint:
        """Called when using as decorator with parenthesis: @Entrypoint()"""
        self.ref = cache_signature_if_function(ref, depth=1)
        return self

    def __repr__(self) -> str:
        return f"<{dotpath(self.__class__)}(ref={self.ref!r})>"

    def __str__(self) -> str:
        field = self.as_graphql_field()
        return undine_settings.SDL_PRINTER.print_field(self.schema_name, field, indent=False)

    def as_graphql_field(self) -> GraphQLField:
        return GraphQLField(
            type_=self.get_field_type(),
            args=self.get_field_arguments(),
            resolve=self.get_resolver(),
            subscribe=self.get_subscription(),
            description=self.description,
            deprecation_reason=self.deprecation_reason,
            extensions=self.extensions,
        )

    def get_field_type(self) -> GraphQLOutputType:
        value = MaybeManyOrNonNull(self.ref, many=self.many, nullable=self.nullable)
        return convert_to_graphql_type(value)  # type: ignore[return-value]

    def get_field_arguments(self) -> GraphQLArgumentMap:
        if self.resolver_func is not None:
            return convert_to_graphql_argument_map(self.resolver_func, many=self.many, entrypoint=True)
        return convert_to_graphql_argument_map(self.ref, many=self.many, entrypoint=True)

    def get_resolver(self) -> GraphQLFieldResolver:
        if self.resolver_func is not None:
            return convert_to_entrypoint_resolver(self.resolver_func, caller=self)
        return convert_to_entrypoint_resolver(self.ref, caller=self)

    def get_subscription(self) -> GraphQLFieldResolver | None:
        return convert_to_entrypoint_subscription(self.ref, caller=self)

    def resolve(self, func: GraphQLFieldResolver | None = None, /) -> GraphQLFieldResolver:
        """
        Decorate a function to add a custom resolver for this Entrypoint.

        >>> class Query(RootType):
        ...     task = Entrypoint(TaskType, many=True)
        ...
        ...     @task.resolve
        ...     def resolve_task(self: Any, info: GQLInfo, name: str) -> list[Task]:
        ...         qs = Task.objects.filter(name__icontains=name)
        ...         return optimize_sync(qs, info)
        """
        if func is None:  # Allow `@<entrypoint_name>.resolve()`
            return self.resolve
        self.resolver_func = cache_signature_if_function(func, depth=1)
        return func

    def permissions(self, func: EntrypointPermFunc | None = None, /) -> EntrypointPermFunc:
        """
        Decorate a function to add it as a permission check for this Entrypoint.

        >>> class Query(RootType):
        ...     task = Entrypoint(TaskType, many=True)
        ...
        ...     @task.permissions
        ...     def task_permissions(self: Any, info: GQLInfo, name: str) -> None:
        ...         raise GraphQLPermissionError
        """
        if func is None:  # Allow `@<entrypoint_name>.permissions()`
            return self.permissions  # type: ignore[return-value]
        self.permissions_func = get_wrapped_func(func)
        return func

    def visible(self, func: VisibilityFunc | None = None, /) -> VisibilityFunc:
        """
        Decorate a function to change the Entrypoint's visibility in the schema.
        Experimental, requires `EXPERIMENTAL_VISIBILITY_CHECKS` to be enabled.

        >>> class Query(RootType):
        ...     task = Entrypoint(TaskType, many=True)
        ...
        ...     @task.visible
        ...     def task_visible(self: Entrypoint, request: DjangoRequestProtocol) -> bool:
        ...         return False
        """
        if func is None:  # Allow `@<entrypoint_name>.visible()`
            return self.visible  # type: ignore[return-value]
        self.visible_func = get_wrapped_func(func)
        return func

    def add_directive(self, directive: Directive, /) -> Self:
        """Add a directive to this entrypoint."""
        check_directives([directive], location=DirectiveLocation.FIELD_DEFINITION)
        self.directives.append(directive)
        return self



from __future__ import annotations

from typing import TYPE_CHECKING

from undine.settings import undine_settings

if TYPE_CHECKING:
    from graphql import (
        GraphQLArgument,
        GraphQLDirective,
        GraphQLEnumType,
        GraphQLEnumValue,
        GraphQLField,
        GraphQLInputField,
        GraphQLInputObjectType,
        GraphQLInterfaceType,
        GraphQLObjectType,
        GraphQLScalarType,
        GraphQLSchema,
        GraphQLUnionType,
    )

    from undine import (
        CalculationArgument,
        Entrypoint,
        Field,
        Filter,
        FilterSet,
        Input,
        InterfaceField,
        InterfaceType,
        MutationType,
        Order,
        OrderSet,
        QueryType,
        RootType,
        UnionType,
    )
    from undine.directives import Directive, DirectiveArgument
    from undine.pagination import OffsetPagination
    from undine.relay import Connection
    from undine.scalars import ScalarType


__all__ = [
    "get_undine_calculation_argument",
    "get_undine_connection",
    "get_undine_directive",
    "get_undine_directive_argument",
    "get_undine_entrypoint",
    "get_undine_field",
    "get_undine_filter",
    "get_undine_filterset",
    "get_undine_input",
    "get_undine_interface_field",
    "get_undine_interface_type",
    "get_undine_mutation_type",
    "get_undine_offset_pagination",
    "get_undine_order",
    "get_undine_orderset",
    "get_undine_query_type",
    "get_undine_root_type",
    "get_undine_scalar",
    "get_undine_schema_directives",
    "get_undine_union_type",
]


def get_undine_schema_directives(schema: GraphQLSchema) -> list[Directive] | None:
    return schema.extensions.get(undine_settings.SCHEMA_DIRECTIVES_EXTENSIONS_KEY)


def get_undine_root_type(object_type: GraphQLObjectType) -> type[RootType] | None:
    return object_type.extensions.get(undine_settings.ROOT_TYPE_EXTENSIONS_KEY)


def get_undine_entrypoint(field: GraphQLField) -> Entrypoint | None:
    return field.extensions.get(undine_settings.ENTRYPOINT_EXTENSIONS_KEY)


def get_undine_query_type(object_type: GraphQLObjectType) -> type[QueryType] | None:
    return object_type.extensions.get(undine_settings.QUERY_TYPE_EXTENSIONS_KEY)


def get_undine_field(field: GraphQLField) -> Field | None:
    return field.extensions.get(undine_settings.FIELD_EXTENSIONS_KEY)


def get_undine_interface_type(object_type: GraphQLInterfaceType) -> type[InterfaceType] | None:
    return object_type.extensions.get(undine_settings.INTERFACE_TYPE_EXTENSIONS_KEY)


def get_undine_interface_field(field: GraphQLField) -> InterfaceField | None:
    return field.extensions.get(undine_settings.INTERFACE_FIELD_EXTENSIONS_KEY)


def get_undine_mutation_type(object_type: GraphQLInputObjectType) -> type[MutationType] | None:
    return object_type.extensions.get(undine_settings.MUTATION_TYPE_EXTENSIONS_KEY)


def get_undine_input(field: GraphQLInputField) -> Input | None:
    return field.extensions.get(undine_settings.INPUT_EXTENSIONS_KEY)


def get_undine_filterset(object_type: GraphQLInputObjectType) -> type[FilterSet] | None:
    return object_type.extensions.get(undine_settings.FILTERSET_EXTENSIONS_KEY)


def get_undine_filter(field: GraphQLInputField) -> Filter | None:
    return field.extensions.get(undine_settings.FILTER_EXTENSIONS_KEY)


def get_undine_orderset(enum: GraphQLEnumType) -> type[OrderSet] | None:
    return enum.extensions.get(undine_settings.ORDERSET_EXTENSIONS_KEY)


def get_undine_order(value: GraphQLEnumValue) -> Order | None:
    return value.extensions.get(undine_settings.ORDER_EXTENSIONS_KEY)


def get_undine_connection(object_type: GraphQLObjectType) -> Connection | None:
    return object_type.extensions.get(undine_settings.CONNECTION_EXTENSIONS_KEY)


def get_undine_offset_pagination(field: GraphQLField) -> OffsetPagination | None:
    return field.extensions.get(undine_settings.OFFSET_PAGINATION_EXTENSIONS_KEY)


def get_undine_directive(directive: GraphQLDirective) -> type[Directive] | None:
    return directive.extensions.get(undine_settings.DIRECTIVE_EXTENSIONS_KEY)


def get_undine_directive_argument(argument: GraphQLArgument) -> DirectiveArgument | None:
    return argument.extensions.get(undine_settings.DIRECTIVE_ARGUMENT_EXTENSIONS_KEY)


def get_undine_scalar(scalar: GraphQLScalarType) -> ScalarType | None:
    return scalar.extensions.get(undine_settings.SCALAR_EXTENSIONS_KEY)


def get_undine_union_type(union: GraphQLUnionType) -> type[UnionType] | None:
    return union.extensions.get(undine_settings.UNION_TYPE_EXTENSIONS_KEY)


def get_undine_calculation_argument(arg: GraphQLArgument) -> CalculationArgument | None:
    return arg.extensions.get(undine_settings.CALCULATION_ARGUMENT_EXTENSIONS_KEY)



from __future__ import annotations

from types import FunctionType
from typing import Any

from django.contrib.contenttypes.fields import GenericForeignKey
from django.db.models import F, Q
from graphql import (
    GraphQLArgument,
    GraphQLArgumentMap,
    GraphQLInputType,
    GraphQLInt,
    GraphQLList,
    GraphQLNonNull,
    GraphQLString,
    GraphQLType,
)

from undine import Calculation, InterfaceField, InterfaceType, MutationType, QueryType, UnionType
from undine.converters import convert_to_graphql_argument_map, convert_to_graphql_type
from undine.dataclasses import LazyGenericForeignKey, LazyLambda, LazyRelation, TypeRef
from undine.exceptions import RegistryMissingTypeError
from undine.pagination import OffsetPagination
from undine.parsers import docstring_parser, parse_is_nullable, parse_parameters
from undine.relay import Connection, Node
from undine.settings import undine_settings
from undine.subscriptions import SignalSubscription
from undine.typing import CombinableExpression, ModelField, RelatedField
from undine.utils.model_utils import get_model_field
from undine.utils.text import get_docstring, to_schema_name

# --- Python types -------------------------------------------------------------------------------------------------


@convert_to_graphql_argument_map.register
def _(ref: FunctionType, **kwargs: Any) -> GraphQLArgumentMap:
    params = parse_parameters(ref)
    docstring = get_docstring(ref)
    arg_descriptions = docstring_parser.parse_arg_descriptions(docstring)
    deprecation_descriptions = docstring_parser.parse_deprecations(docstring)

    arguments: GraphQLArgumentMap = {}
    kwargs["is_input"] = True
    for param in params:
        graphql_type = convert_to_graphql_type(param.annotation, **kwargs)
        nullable = parse_is_nullable(param.annotation)
        if not nullable:
            graphql_type = GraphQLNonNull(graphql_type)

        arguments[to_schema_name(param.name)] = GraphQLArgument(
            graphql_type,
            default_value=param.default_value,
            description=arg_descriptions.get(param.name),
            deprecation_reason=deprecation_descriptions.get(param.name),
            out_name=param.name,
        )

    return arguments


# --- Model fields -------------------------------------------------------------------------------------------------


@convert_to_graphql_argument_map.register
def _(_: ModelField | CombinableExpression | F | Q, **kwargs: Any) -> GraphQLArgumentMap:
    return {}


@convert_to_graphql_argument_map.register
def _(_: RelatedField, **kwargs: Any) -> GraphQLArgumentMap:
    return {}


@convert_to_graphql_argument_map.register  # Required for Django<5.1
def _(_: GenericForeignKey, **kwargs: Any) -> GraphQLArgumentMap:  # pragma: no cover
    return {}


# --- Custom types -------------------------------------------------------------------------------------------------


@convert_to_graphql_argument_map.register
def _(ref: type[QueryType], **kwargs: Any) -> GraphQLArgumentMap:
    if not kwargs["many"]:
        if not kwargs.get("entrypoint"):
            return {}

        field = get_model_field(model=ref.__model__, lookup="pk")
        input_type = convert_to_graphql_type(field, model=ref.__model__)
        input_type = GraphQLNonNull(input_type)
        return {"pk": GraphQLArgument(input_type, out_name="pk")}

    arguments: GraphQLArgumentMap = {}

    if ref.__filterset__:
        input_type = ref.__filterset__.__input_type__()
        arguments[undine_settings.QUERY_TYPE_FILTER_INPUT_KEY] = GraphQLArgument(input_type)

    if ref.__orderset__:
        enum_type = ref.__orderset__.__enum_type__()
        input_type = GraphQLList(GraphQLNonNull(enum_type))
        arguments[undine_settings.QUERY_TYPE_ORDER_INPUT_KEY] = GraphQLArgument(input_type)

    return arguments


@convert_to_graphql_argument_map.register
def _(ref: type[MutationType], **kwargs: Any) -> GraphQLArgumentMap:
    input_type = ref.__input_type__()

    arguments: GraphQLArgumentMap = {}

    arg_type: GraphQLInputType = GraphQLNonNull(input_type)
    if kwargs["many"]:
        arg_type = GraphQLNonNull(GraphQLList(arg_type))

    arguments[undine_settings.MUTATION_INPUT_DATA_KEY] = GraphQLArgument(arg_type)
    return arguments


@convert_to_graphql_argument_map.register
def _(ref: type[UnionType], **kwargs: Any) -> GraphQLArgumentMap:
    kwargs["many"] = True

    arguments: GraphQLArgumentMap = {}

    for model, query_type in ref.__query_types_by_model__.items():
        args = convert_to_graphql_argument_map(query_type, **kwargs)

        if undine_settings.QUERY_TYPE_FILTER_INPUT_KEY in args:
            filter_key = f"{undine_settings.QUERY_TYPE_FILTER_INPUT_KEY}{model.__name__}"
            arguments[filter_key] = args[undine_settings.QUERY_TYPE_FILTER_INPUT_KEY]

        if undine_settings.QUERY_TYPE_ORDER_INPUT_KEY in args:
            order_by_key = f"{undine_settings.QUERY_TYPE_ORDER_INPUT_KEY}{model.__name__}"
            arguments[order_by_key] = args[undine_settings.QUERY_TYPE_ORDER_INPUT_KEY]

    if ref.__filterset__:
        input_type = ref.__filterset__.__input_type__()
        arguments[undine_settings.QUERY_TYPE_FILTER_INPUT_KEY] = GraphQLArgument(input_type)

    if ref.__orderset__:
        enum_type = ref.__orderset__.__enum_type__()
        input_type = GraphQLList(GraphQLNonNull(enum_type))
        arguments[undine_settings.QUERY_TYPE_ORDER_INPUT_KEY] = GraphQLArgument(input_type)

    return arguments


@convert_to_graphql_argument_map.register
def _(ref: LazyRelation, **kwargs: Any) -> GraphQLArgumentMap:
    try:
        value = ref.get_type()
    except RegistryMissingTypeError:
        value = ref.field

    return convert_to_graphql_argument_map(value, **kwargs)


@convert_to_graphql_argument_map.register
def _(_: LazyGenericForeignKey, **kwargs: Any) -> GraphQLArgumentMap:
    return {}


@convert_to_graphql_argument_map.register
def _(ref: LazyLambda, **kwargs: Any) -> GraphQLArgumentMap:
    return convert_to_graphql_argument_map(ref.callback(), **kwargs)


@convert_to_graphql_argument_map.register
def _(_: TypeRef, **kwargs: Any) -> GraphQLArgumentMap:
    return {}


@convert_to_graphql_argument_map.register
def _(ref: type[Calculation], **kwargs: Any) -> GraphQLArgumentMap:
    arguments: GraphQLArgumentMap = {}

    for arg in ref.__arguments__.values():
        arguments[arg.schema_name] = arg.as_graphql_argument()

    return arguments


@convert_to_graphql_argument_map.register
def _(ref: Connection, **kwargs: Any) -> GraphQLArgumentMap:
    kwargs["many"] = True

    if ref.union_type is not None:
        arguments = convert_to_graphql_argument_map(ref.union_type, **kwargs)
    elif ref.interface_type is not None:
        arguments = convert_to_graphql_argument_map(ref.interface_type, **kwargs)
    else:
        arguments = convert_to_graphql_argument_map(ref.query_type, **kwargs)

    return {
        "after": GraphQLArgument(
            GraphQLString,
            description="Only return items in the connection that come after this cursor.",
            out_name="after",
        ),
        "before": GraphQLArgument(
            GraphQLString,
            description="Only return items in the connection that come before this cursor.",
            out_name="before",
        ),
        "first": GraphQLArgument(
            GraphQLInt,
            description="Number of items to return from the start.",
            out_name="first",
        ),
        "last": GraphQLArgument(
            GraphQLInt,
            description="Number of items to return from the end (after evaluating first).",
            out_name="last",
        ),
        **arguments,
    }


@convert_to_graphql_argument_map.register
def _(ref: OffsetPagination, **kwargs: Any) -> GraphQLArgumentMap:
    kwargs["many"] = True

    if ref.union_type is not None:
        arguments = convert_to_graphql_argument_map(ref.union_type, **kwargs)
    elif ref.interface_type is not None:
        arguments = convert_to_graphql_argument_map(ref.interface_type, **kwargs)
    else:
        arguments = convert_to_graphql_argument_map(ref.query_type, **kwargs)

    return {
        "offset": GraphQLArgument(
            GraphQLInt,
            description="Number of items to skip from the start.",
            out_name="offset",
        ),
        "limit": GraphQLArgument(
            GraphQLInt,
            description="Maximum number of items to return from the start (after applying offset).",
            out_name="limit",
        ),
        **arguments,
    }


@convert_to_graphql_argument_map.register
def _(ref: type[InterfaceType], **kwargs: Any) -> GraphQLArgumentMap:
    kwargs["many"] = True

    arguments: GraphQLArgumentMap = {}

    for query_type in ref.__concrete_implementations__():
        model = query_type.__model__
        args = convert_to_graphql_argument_map(query_type, **kwargs)

        if undine_settings.QUERY_TYPE_FILTER_INPUT_KEY in args:
            filter_key = f"{undine_settings.QUERY_TYPE_FILTER_INPUT_KEY}{model.__name__}"
            arguments[filter_key] = args[undine_settings.QUERY_TYPE_FILTER_INPUT_KEY]

        if undine_settings.QUERY_TYPE_ORDER_INPUT_KEY in args:
            order_by_key = f"{undine_settings.QUERY_TYPE_ORDER_INPUT_KEY}{model.__name__}"
            arguments[order_by_key] = args[undine_settings.QUERY_TYPE_ORDER_INPUT_KEY]

    return arguments


@convert_to_graphql_argument_map.register
def _(ref: InterfaceField, **kwargs: Any) -> GraphQLArgumentMap:
    return ref.args


@convert_to_graphql_argument_map.register
def _(ref: type[Node], **kwargs: Any) -> GraphQLArgumentMap:
    return {
        ref.id.schema_name: GraphQLArgument(
            ref.id.output_type,  # type: ignore[arg-type]
            description=ref.id.description,
            out_name=ref.id.name,
        ),
    }


@convert_to_graphql_argument_map.register
def _(_: SignalSubscription, **kwargs: Any) -> GraphQLArgumentMap:
    return {}


# --- GraphQL types ------------------------------------------------------------------------------------------------


@convert_to_graphql_argument_map.register
def _(_: GraphQLType, **kwargs: Any) -> GraphQLArgumentMap:
    return {}
